@@ -3,6 +3,7 @@ import subprocess
 import os
 import argparse
 import lmql
+from lmql.runtime.interpreter import LMQLResult
 
 import lmql.version as version_info
 
@@ -18,40 +19,51 @@ def cmd_run():
     emoji:🏃 run a LMQL script (e.g. "lmql run latest/hello.lmql")
     """
     import asyncio
+    import time
+
+    start = time.time()
 
     parser = argparse.ArgumentParser(description="Runs a LMQL program.")
     parser.add_argument("lmql_file", type=str, help="path to the LMQL file to run")
     parser.add_argument("--no-clear", action="store_true", dest="no_clear", help="don't clear inbetween printing results")
     parser.add_argument("--no-realtime", action="store_true", dest="no_realtime", help="don't print text as it's being generated")
+    parser.add_argument("--time", action="store_true", dest="time", help="Time the query.")
 
     args = parser.parse_args(sys.argv[2:])
 
     absolute_path = os.path.abspath(args.lmql_file)
-    from lmql.runtime.output_writer import PrintingDebuggerOutputWriter
 
-    writer = PrintingDebuggerOutputWriter()
+    writer = lmql.printing
     writer.clear = not args.no_clear
     writer.print_output = not args.no_realtime
 
-    results = asyncio.run(lmql.run_file(absolute_path, output_writer=writer))
+    if os.path.exists(absolute_path):
+        results = asyncio.run(lmql.run_file(absolute_path, output_writer=writer))
+    else:
+        code = args.lmql_file
+        results = asyncio.run(lmql.run(code, output_writer=writer))
     
     if type(results) is not list:
         results = [results]
     
     for r in results:
-        for v in [v for v in r.variables if v.startswith("P(")]:
-            distribution = r.variables[v]
-            max_prob = max([p for _,p in distribution])
-            labels = []
-            for value, prob in distribution:
-                label = value if prob != max_prob else f"{value} (*)"
-                labels.append(label)
-            max_length = max([len(str(l)) for l in labels])
+        if isinstance(r, LMQLResult):
+            for v in [v for v in r.variables if v.startswith("P(")]:
+                distribution = r.variables[v]
+                max_prob = max([p for _,p in distribution])
+                labels = []
+                for value, prob in distribution:
+                    label = value if prob != max_prob else f"{value} (*)"
+                    labels.append(label)
+                max_length = max([len(str(l)) for l in labels])
 
-            print(v)
-            for (value, prob), label in zip(distribution, labels):
-                label = label.ljust(max_length)
-                print(" - {} {}".format(label, prob))
+                print(v)
+                for (value, prob), label in zip(distribution, labels):
+                    label = label.ljust(max_length)
+                    print(" - {} {}".format(label, prob))
+
+    if args.time:
+        print("Query took:", time.time() - start)
 
 def ensure_node_install():
     try:
@@ -150,20 +162,36 @@ def cmd_version():
 
 def hello():
     import asyncio
-    code_local = """
-argmax "Hello[WHO]" from "local:gpt2-medium" where len(WHO) < 10    
-"""
-    print("[Greeting 🤗 Transformers]")
-    asyncio.run(lmql.run(code_local))
+    backend = None
+    # check for additional arg
+    if len(sys.argv) > 2:
+        backend = sys.argv[2]
+        if backend not in ["hf", "openai"]:
+            print("Invalid backend, please specify one of {}".format(", ".join(["hf", "openai"])))
+            sys.exit(1)
     
-    print("[Greeting OpenAI]")
-    code_openai = """
-argmax "Hello[WHO]" from "openai/text-ada-001" where len(WHO) < 10    
-"""
-    asyncio.run(lmql.run(code_openai))
+    if backend is None or backend == "hf":
+        code_local = """
+    argmax "Hello[WHO]" from "local:gpt2-medium" where len(WHO) < 10    
+    """
+        print("[Greeting 🤗 Transformers]")
+        asyncio.run(lmql.run(code_local, output_writer=lmql.printing))
+    
+    if backend is None or backend == "openai":
+        print("[Greeting OpenAI]")
+        code_openai = """
+    argmax "Hello[WHO]" from "openai/text-ada-001" where len(WHO) < 10    
+    """
+        asyncio.run(lmql.run(code_openai, output_writer=lmql.printing))
+
+def basic_samples():
+    from lmql.tests.test_sample_queries import main
+    import asyncio
+    asyncio.run(main())
 
 hidden_commands = {
-    "hello": hello
+    "hello": hello,
+    "test": basic_samples
 }
 
 def main():
@@ -190,4 +218,4 @@ def main():
     command[0]()
 
 if __name__ == "__main__":
-    main()
\ No newline at end of file
+    main()
