@@ -1,5 +1,7 @@
+from _ast import AsyncFunctionDef, ClassDef, FunctionDef, Import, ImportFrom
 import ast
 import sys
+from typing import Any
 import astunparse
 import os
 import importlib
@@ -12,18 +14,99 @@ from lmql.language.qstrings import qstring_to_stmts, TemplateVariable
 from lmql.language.validator import LMQLValidator, LMQLValidationError
 from lmql.language.fragment_parser import LMQLDecoderConfiguration, LMQLQuery, LanguageFragmentParser, FragmentParserError
 from lmql.runtime.model_registry import model_name_aliases
+import lmql.runtime.lmql_runtime as lmql_runtime
+from lmql.runtime.dclib import get_all_decoders
+
+OPS_NAMESPACE = "lmql.ops"
+
+class FreeVarCollector(ast.NodeVisitor):
+    def __init__(self, free_vars, exclude_criteria=None):
+        self.free_vars = free_vars
+        
+        self.exclude = exclude_criteria or []
+
+    def is_excluded(self, name):
+        for criterion in self.exclude:
+            if type(criterion) is list:
+                return name in criterion
+            elif callable(criterion):
+                return criterion(name)
+            else:
+                assert False, f"compiler error: invalid type for exclusion criterion in identifier collection: {type(criterion)}"
+        return False
+
+    def visit_Name(self, node):
+        if type(node.ctx) is ast.Load:
+            if self.is_excluded(node.id):
+                return
+            self.free_vars.add(node.id)
+
+class DefinedVarsCollector(ast.NodeVisitor):
+    def __init__(self, defined_vars):
+        self.defined_vars = defined_vars
+
+    def visit_Name(self, node):
+        if type(node.ctx) is ast.Store:
+            self.defined_vars.add(node.id)
+
+    def visit_FunctionDef(self, node: FunctionDef) -> Any:
+        self.defined_vars.add(node.name)
+    
+    def visit_ClassDef(self, node: ClassDef) -> Any:
+        self.defined_vars.add(node.name)
+
+    def visit_Import(self, node: Import) -> Any:
+        for alias in node.names:
+            self.defined_vars.add(alias.asname or alias.name)
+
+    def visit_ImportFrom(self, node: ImportFrom) -> Any:
+        for alias in node.names:
+            self.defined_vars.add(alias.asname or alias.name)
+
+    def visit_AsyncFunctionDef(self, node: AsyncFunctionDef) -> Any:
+        self.defined_vars.add(node.name)
 
 class PromptScope(ast.NodeVisitor):
+    def __init__(self):
+        self.distribution_vars = set()
+        self.defined_vars = set()
+        self.prologue_vars = set()
+        self.free_vars = set()
+        self.written_vars = set()
+
+    def scope_prologue(self, query: LMQLQuery):
+        if query.prologue is None: return
+        
+        DefinedVarsCollector(self.prologue_vars).visit(ast.parse(query.prologue))
+
     def scope(self, query: LMQLQuery):
         self.distribution_vars = set([query.distribution.variable_name]) if query.distribution is not None else set()    
         self.defined_vars = set()
+        self.prologue_vars = set()
 
         # collect set of global query template variables
         self.free_vars = set()
         self.written_vars = set()
 
+        # collect defined vars in prologue
+        self.scope_prologue(query)
+
+        # collect defined vars in prompt
         for p in query.prompt: self.visit(p)
 
+        # also collect variable reads from where clause
+        if query.where is not None:
+            FreeVarCollector(self.free_vars, exclude_criteria=[self.exclude_identifier]).visit(query.where)
+        if query.from_ast is not None:
+            FreeVarCollector(self.free_vars, exclude_criteria=[self.exclude_identifier]).visit(query.from_ast)
+        if query.decode is not None:
+            FreeVarCollector(self.free_vars, exclude_criteria=[self.exclude_identifier]).visit(query.decode)
+        if query.distribution is not None:
+            FreeVarCollector(self.free_vars, exclude_criteria=[self.exclude_identifier]).visit(query.distribution.values)
+
+        # remove all defined and prologue vars from free vars
+        self.free_vars = self.free_vars - self.defined_vars - self.prologue_vars
+
         query.scope = self
 
     def visit_Constant(self, node):
@@ -37,16 +120,21 @@ class PromptScope(ast.NodeVisitor):
             self.written_vars.add(v)
             if v in self.free_vars: self.free_vars.remove(v)
 
-        # capture set of format vars
-        used_fstring_expr = [v[1:-1] for v in re.findall("\{[^\}]+\}", qstring)]
+        # capture set of format vars (exclude {{ and }})
+        # replace {{ and }} with escape sequence \u007b and \u007d 
+        qstring = qstring.replace("{{", "__curly_open__").replace("}}", "__curly_close__")
+        used_fstring_expr = [v[1:-1] for v in re.findall("\{[^\}\{]+\}", qstring)]
         for v in used_fstring_expr:
             if v.startswith(":"):
                 continue
-            parsed = ast.parse(v).body[0].value
-            self.visit(parsed)
-            
-            # if v not in self.defined_vars and v not in self.free_vars and v not in self.written_vars:
-            #     self.free_vars.add(v)
+            try:
+                parsed = ast.parse(v).body[0].value
+                self.visit(parsed)
+            except:
+                raise RuntimeError("Failed to parse fstring expression: ", v)
+
+        # put double curly braces back in
+        qstring = qstring.replace("__curly_open__", "{{").replace("__curly_close__", "}}")
                 
         template_tags = [v[1:-1] for v in re.findall("\{:[A-z0-9]+\}", qstring)]
         for tt in template_tags:
@@ -56,6 +144,24 @@ class PromptScope(ast.NodeVisitor):
 
         return super().visit_Constant(node)
 
+    def exclude_identifier(self, name):
+        # make sure "input" can be intercepted
+        if name in ["input"]:
+            return False
+        if name in self.free_vars:
+            return True
+        if name in self.written_vars:
+            return True
+        # exclude LMQL built-ins
+        if get_builtin_name(name) is not None:
+            return True
+        # exclude Python built-ins
+        if __builtins__.get(name) is not None:
+            return True
+        if name in get_all_decoders():
+            return True
+        return False
+
     def visit_Name(self, node: ast.Name):
         name = str(node.id)
         
@@ -65,8 +171,9 @@ class PromptScope(ast.NodeVisitor):
                 self.free_vars.remove(name)
             
         if type(node.ctx) is ast.Load:
-            if name not in self.free_vars and name not in self.written_vars:
-                self.free_vars.add(name)
+            if self.exclude_identifier(name):
+                return
+            self.free_vars.add(name)
         
         return True
 
@@ -87,6 +194,16 @@ class QueryStringTransformation(ast.NodeTransformer):
         else:
             self.generic_visit(expr)
         return expr
+    
+    # translate id access to context
+    def visit_Name(self, node: ast.Name):
+        name = str(node.id)
+        
+        if type(node.ctx) is ast.Load:
+            if name == "context":
+                return ast.parse("(" + yield_call("get_context", ()) + ")").body[0].value
+        return node
+
 
     def transform_Constant(self, constant):
         if type(constant.value) is not str: return constant
@@ -107,10 +224,15 @@ class QueryStringTransformation(ast.NodeTransformer):
                     declared_template_vars.add(stmt.name)
                     compiled_qstring += "[" + stmt.name + "]"
 
-        result_code = f'yield context.query(f"""{compiled_qstring}""")'
+        if len(compiled_qstring) == 0:
+            return constant
+
+        # result_code = f'yield context.query(f"""{compiled_qstring}""")'
+        result_code = interrupt_call('query', f'f"""{compiled_qstring}"""')
 
         for v in declared_template_vars:
-            result_code += f"\n{v} = context.get_var('{v}')"
+            get_var_call = yield_call('get_var', f'"{v}"')
+            result_code += f"\n{v} = " + get_var_call
         return ast.parse(result_code)
 
     # def transform_prompt_stmt(self, stmt):
@@ -198,7 +320,7 @@ class WhereClauseTransformation():
 
         # check for template variables
         if node.id in self.scope.defined_vars and not keep_variables:
-            return f"lmql.Var('{node.id}')"
+            return f"lmql.runtime_support.Var('{node.id}')"
 
         return bn or node.id
 
@@ -209,7 +331,7 @@ class WhereClauseTransformation():
                 tops = [self.transform_node(op, snf) for op in ops]
                 tops_list = ",\n  ".join([t.strip() or "None" for t in tops])
                 
-                Op = "lmql.AndOp" if type(expr.op) is ast.And else "lmql.OrOp"
+                Op = f"{OPS_NAMESPACE}.AndOp" if type(expr.op) is ast.And else f"{OPS_NAMESPACE}.OrOp"
                 return snf.add(f"{Op}([\n  {tops_list}\n])")
         # elif type(expr) is ast.Call:
         #     tfunc = self.transform_node(expr.func, snf)
@@ -222,7 +344,7 @@ class WhereClauseTransformation():
             op = expr.op
             
             Ops = {
-                ast.Not: "lmql.NotOp"
+                ast.Not: f"{OPS_NAMESPACE}.NotOp"
             }
 
             for OpT, impl in Ops.items():
@@ -236,10 +358,10 @@ class WhereClauseTransformation():
             assert len(expr.ops) == 1, "compiler currently does not support comparison with more than one operator"
             
             Ops = {
-                ast.Eq: "lmql.EqOp",
-                ast.Lt: "lmql.Lt",
-                ast.Gt: "lmql.Gt",
-                ast.In: "lmql.InOp"
+                ast.Eq: f"{OPS_NAMESPACE}.EqOp",
+                ast.Lt: f"{OPS_NAMESPACE}.Lt",
+                ast.Gt: f"{OPS_NAMESPACE}.Gt",
+                ast.In: f"{OPS_NAMESPACE}.InOp"
             }
 
             for OpT, impl in Ops.items():
@@ -248,6 +370,11 @@ class WhereClauseTransformation():
                     ops_list = ", ".join(ops).strip()
                     return snf.add(f"{impl}([{ops_list}])")
             
+            # if is_type_constraint(expr):
+            #     type_name = expr.comparators[0].id
+            #     var_name = expr.left.args[0].id
+            #     return snf.add(f"{OPS_NAMESPACE}.TypeConstraint([{type_name}, {OPS_NAMESPACE}.Var('{var_name}')])")
+            
             assert False, "operator {} is not supported.".format(astunparse.unparse(expr))
         elif type(expr) is ast.Constant:
             return self.default_transform_node(expr, snf).strip()
@@ -286,10 +413,10 @@ class WhereClauseTransformation():
         node = NameTransformer(transform_name).visit(node)
 
         args = (" " + ", ".join(names)) if len(names) > 0 else ""
-        var_ops = (", ".join([f"lmql.Var('{n}')" for n in names])).strip() if len(names) > 0 else ""
+        var_ops = (", ".join([f"{OPS_NAMESPACE}.Var('{n}')" for n in names])).strip() if len(names) > 0 else ""
         fct_code = astunparse.unparse(node).strip()
         
-        return f"lmql.OpaqueLambdaOp([lambda{args}: {fct_code}, {var_ops}])"
+        return f"{OPS_NAMESPACE}.OpaqueLambdaOp([lambda{args}: {fct_code}, {var_ops}])"
 
 def is_allowed_builtin_python_call(node):
     if type(node) is not ast.Name:
@@ -298,15 +425,39 @@ def is_allowed_builtin_python_call(node):
     return node.id in allowed_builtin_functions
 
 def get_builtin_name(node, plain_python=False):
-    if type(node) is not ast.Name:
-        return None
-    n = node.id
+    if type(node) is str:
+        n = node
+    else:
+        if type(node) is not ast.Name:
+            return None
+        n = node.id
     
     if n in lmql_operation_registry.keys():
-        return lmql_operation_registry[n]
+        return OPS_NAMESPACE + "." + lmql_operation_registry[n]
     
     return None
 
+def is_type_constraint(expr: ast.Expr):
+    if not type(expr.left) is ast.Call:
+        return False
+    if not type(expr.left.func) is ast.Name:
+        return False
+    if not expr.left.func.id == "type":
+        return False
+    if not type(expr.ops[0]) is ast.Is:
+        return False
+    if not len(expr.comparators) == 1:
+        return False
+    right = expr.comparators[0]
+    if not type(right) is ast.Name:
+        return False
+    if not len(expr.left.args) == 1:
+        return False
+    if not type(expr.left.args[0]) is ast.Name:
+        return False
+    
+    return True
+
 class DecodeClauseTransformation:
     def __init__(self, query):
         self.query = query
@@ -352,7 +503,7 @@ class PythonFunctionWriter:
         self.file = open(self.filename, "w")
 
         self.indent = "  "
-        self.write("import lmql.runtime.lmql_runtime as lmql\n")
+        self.write("import lmql\n")
         self.write(self.prologue)
         
         if decorators is not None:
@@ -394,6 +545,9 @@ class LMQLModule(object):
     def load(self):
         sys.path.append(os.path.dirname(self.compiled_file))
         m = __import__(os.path.basename(self.compiled_file[:-3]))
+        for v in m.__dict__.values():
+            if type(v) is lmql_runtime.LMQLQueryFunction:
+                v.lmql_code = self.lmql_code
         setattr(m, "code", self.code)
         setattr(m, "lmql_code", self.lmql_code)
         return m
@@ -426,6 +580,12 @@ def preprocess_text(lmql_code):
     common_indent = min([len(l) - len(l.lstrip()) for l in lines if len(l.strip()) > 0])
     return "\n".join([l[common_indent:] for l in lines])
 
+def yield_call(func, *args):
+    return f"""yield lmql.runtime_support.context_call("{func}", {", ".join([str(a) for a in args])})"""
+
+def interrupt_call(func, *args):
+    return f"""yield lmql.runtime_support.interrupt_call("{func}", {", ".join([str(a) for a in args])})"""
+
 class LMQLCompiler:
     def __init__(self):
         pass
@@ -465,21 +625,25 @@ class LMQLCompiler:
             output_variables = "output_variables=[" + ", ".join([f'"{v}"' for v in scope.defined_vars]) + "]"
 
             # generate function that runs query
-            with PythonFunctionWriter("query", output_file, list(scope.free_vars) + ["context"], 
-                q.prologue, decorators=["lmql.query"], decorators_args=[output_variables]) as writer:
+            parameters = list(sorted(list(scope.free_vars.union(set(["context"])))))
+
+            with PythonFunctionWriter("query", output_file, parameters, 
+                q.prologue, decorators=["lmql.compiled_query"], decorators_args=[output_variables]) as writer:
                 
-                writer.add(f"context.set_model({model_name})")
-                writer.add(f"context.set_decoder({astunparse.unparse(q.decode.method).strip()}, {unparse_list(q.decode.decoding_args)})")
+                writer.add(yield_call("set_model", model_name))
+                # writer.add(f"context.set_decoder({})")
+                writer.add(yield_call("set_decoder", astunparse.unparse(q.decode.method).strip(), unparse_list(q.decode.decoding_args)))
                 writer.add("# where")
                 writer.add(q.where)
-                writer.add(f"context.set_where_clause({q.where_expr})")
+                writer.add(yield_call("set_where_clause", q.where_expr))
                 writer.add("# prompt")
                 writer.add(astunparse.unparse(q.prompt))
                 if q.distribution:
                     writer.add("# distribution")
-                    writer.add("context.set_distribution('{}', {})".format(q.distribution.variable_name, astunparse.unparse(q.distribution.values).strip()))
+                    # writer.add("context.set_distribution('{}', {})".format(q.distribution.variable_name, astunparse.unparse(q.distribution.values).strip()))
+                    writer.add(yield_call("set_distribution", "\"" + q.distribution.variable_name + "\"", astunparse.unparse(q.distribution.values).strip()))
                 
-                writer.add(f"yield ('result', context.get_return_value())")
+                writer.add(f"yield ('result', (" + yield_call("get_return_value", ()) + "))")
 
             return LMQLModule(output_file, lmql_code=lmql_code, output_variables=[v for v in scope.defined_vars])
         except FragmentParserError as e:
