@@ -11,9 +11,16 @@ from functools import total_ordering
 
 from .openai_api import complete, OpenAIRateLimitError, Capacity
 
-class EmptyStreamError(Exception): pass
+global logit_bias_logging
+logit_bias_logging = True
 
+def set_logit_bias_logging(value):
+    global logit_bias_logging
+    logit_bias_logging = value
+
+class EmptyStreamError(Exception): pass
 class ChaosException(openai.APIError): pass
+class APIShutDownException(RuntimeError): pass
 class MaximumRetriesExceeded(Exception): 
     def __init__(self, error: Exception, retries: int):
         self.error = error
@@ -175,18 +182,7 @@ class ResponseStream:
                     self.stats.tokens += len(c["logprobs"]["tokens"])
                     assert c is not None
                     self.slices[index].digest(c)
-                    
-                    if c["finish_reason"] == "stop":
-                        # "<|endoftext|>", "<|endoftext|>", 0.0
-                        self.slices[index].digest({
-                            "text": "<|endoftext|>",
-                            "logprobs": {
-                                "text_offset": [0],
-                                "token_logprobs": [0.0],
-                                "tokens": ["<|endoftext|>"],
-                                "top_logprobs": [{"<|endoftext|>": 0.0}]
-                            }
-                        })
+                    self.slices[index].finish_reason = c["finish_reason"]
                     
                     # logprobs.tokens, text, logprobs.token_logprobs
             for c in self.slices:
@@ -356,6 +352,7 @@ class ResponseStreamSliceIterator:
         self.retries = 0
         self.text = ""
         self.consumed_tokens = []
+        self.n = 0
 
     async def recover(self):
         recovery_kwargs = self.slice.kwargs.copy()
@@ -400,22 +397,71 @@ class ResponseStreamSliceIterator:
         # otherwise the chunking aligns with the old stream, so we return the next chunk
         return await self.__anext__()
 
+    async def get_next(self):
+        if self.slice.done.is_set(): 
+            if self.n == 0:
+                return RecoveryAttempt(self.slice.kwargs, TimeoutError(), self.slice.maximum_retries)
+            raise StopAsyncIteration
+        check_done_task = asyncio.create_task(self.slice.done.wait())
+        get_next_item_task = asyncio.create_task(self.slice.data_queue.get())
+        done, pending = await asyncio.wait([get_next_item_task, check_done_task], 
+            return_when=asyncio.FIRST_COMPLETED, timeout=2.0)
+        
+
+        if check_done_task in done:
+            # this indicates the end of this response stream
+            for t in pending: t.cancel()
+            if self.n == 0:
+                return RecoveryAttempt(self.slice.kwargs, TimeoutError(), self.slice.maximum_retries)
+            raise StopAsyncIteration
+        elif len(done) > 0:
+            assert get_next_item_task in done, f"expected get_next_item_task to be done, but only {done} is done."
+            # cancel self.done waiting task
+            for t in pending: t.cancel()
+            # return with new data chunk
+            self.n += 1
+            return get_next_item_task.result()
+        else:
+            for t in pending: t.cancel()
+            # if after timeout this response has been fully consumed, we are done
+            if self.slice.done.is_set() and self.n > 0:
+                raise StopAsyncIteration
+            # otherwise return a RecoveryAttempt for retrying this request
+            return RecoveryAttempt(self.slice.kwargs, TimeoutError(), self.slice.maximum_retries)
+
     async def __anext__(self):
         try:
-            try:
-                data = await asyncio.wait_for(self.slice.data_queue.get(), 2.0)
-            except asyncio.TimeoutError as e:
-                data = RecoveryAttempt(self.slice.kwargs, e, self.slice.maximum_retries)
+            data = await self.get_next()
             # None indicates end of stream
-            if data is None: raise StopAsyncIteration
+            if data is None:
+                if self.slice.done.is_set():
+                    raise StopAsyncIteration
+                else:
+                    if self.slice.finish_reason != "length":
+                        # return eos token as last item, if stream did not finish due to length
+                        data = {
+                            "text": "<|endoftext|>",
+                            "logprobs": {
+                                "text_offset": [0],
+                                "token_logprobs": [0.0],
+                                "tokens": ["<|endoftext|>"],
+                                "top_logprobs": [{"<|endoftext|>": 0.0}]
+                            }
+                        }
+                        self.slice.done.set()
+                    else:
+                        self.slice.done.set()
+                        raise StopAsyncIteration
             # exceptions that are queued are definitive (all retries failed)
             if isinstance(data, Exception): raise data
             # RecoveryAttempt indicates that the underlying stream errored out and we need to recover (still retries left)
             if isinstance(data, RecoveryAttempt):
+                if not self.slice.stream.scheduler.is_available():
+                    # fail quietly, if parent scheduler is no longer available (results of this query will be discarded anyway)
+                    raise StopAsyncIteration()
                 # if the stream of our self.slice errors out, we can recover by creating a new 
                 # stream via a new call to openai.Completion.create
                 attempt: RecoveryAttempt = data
-                traceback.print_exc()
                 print("OpenAI API: Underlying stream of OpenAI complete() call failed with error", type(attempt.error), attempt.error, f"Retrying... (attempt: {self.retries})", flush=True)
                 self.retries += 1
                 # if we have exceeded the maximum number of retries, raise the error
@@ -433,7 +479,6 @@ class ResponseStreamSliceIterator:
         except asyncio.CancelledError:
             raise StopAsyncIteration
 
-
 class ResponseStreamSlice:
     def __init__(self, stream, kwargs, maximum_retries=3):
         self.stream: ResponseStream = stream
@@ -442,11 +487,15 @@ class ResponseStreamSlice:
 
         self.data_queue = asyncio.Queue()
         self.failed = False
+        self.done = asyncio.Event()
+        self.finish_reason = None
+
+        self.itr = None
 
     def digest(self, data):
         assert not self.failed, f"digest called on failed slice"
         self.data_queue.put_nowait(data)
-    
+
     def finish(self):
         assert not self.failed, f"finish called on failed slice"
         self.data_queue.put_nowait(None)
@@ -510,6 +559,11 @@ class AsyncOpenAIAPI:
         self.futures = set()
         self.restore_cache()
 
+        self.first_token_latency = 0
+
+    def reset_latency_stats(self):
+        self.first_token_latency = 0
+
     def restore_cache(self):
         if not self.use_cache:
             return
@@ -550,9 +604,12 @@ class AsyncOpenAIAPI:
         self.complete_api_worker.cancel()
         for worker in self.complete_request_workers:
             worker.cancel()
-        loop = asyncio.get_event_loop()
-        while not all([t.done() for t in (self.complete_request_workers + [self.complete_api_worker])]):
-            loop._run_once()
+        try:
+            loop = asyncio.get_event_loop()
+            while not all([t.done() for t in (self.complete_request_workers + [self.complete_api_worker])]):
+                loop._run_once()
+        except:
+            pass # if no more event loop is around, no need to wait for the workers to finish
 
     async def api_complete_worker(self, queue):
         while True:
@@ -579,8 +636,6 @@ class AsyncOpenAIAPI:
         res = complete(**kwargs)
         first = await anext(res)
         return first_buffered(res, first)
-        
-        return res
 
     def is_definitive_error(self, e):
         if "logit biases, but can provide at most" in str(e):
@@ -614,8 +669,8 @@ class AsyncOpenAIAPI:
                             t = (2.0 * random.random()) ** (self.maximum_retries - retries)
                             print("Backing off for", t , "seconds")
                             await asyncio.sleep(t)
-            except asyncio.CancelledError:
-                return
+            except asyncio.CancelledError as e:
+                break
             except Exception as e:
                 print("error", type(e))
                 for future in futures:
@@ -650,7 +705,9 @@ class AsyncOpenAIAPI:
                 biases = biases[:299] + [(50256, kwargs["logit_bias"][50256])]
             else:
                 biases = biases[:300]
-            print("warning: the required logit_bias is too large to be handled by the OpenAI API and will be limited to the first 300 tokens. This can lead to the violation of the provided constraints or undesired model output. To avoid this use less broad or no constraints.")
+            global logit_bias_logging
+            if logit_bias_logging:
+                print("warning: the required logit_bias is too large to be handled by the OpenAI API and will be limited to the first 300 tokens. This can lead to the violation of the provided constraints or undesired model output. To avoid this use less broad or no constraints.", file=sys.stderr)
             kwargs["logit_bias"] = {t:b for t,b in biases}
 
 
@@ -659,5 +716,9 @@ class AsyncOpenAIAPI:
         r = RequestQueueItem(kwargs, request_id)
         await self.complete_api_call_queue.put(r)
         self.request_ctr += 1
-        
-        return await result_fut
\ No newline at end of file
+        if not self.is_available():
+            raise APIShutDownException(f"bopenai requires at least one worker to be running to issue new complete requests.")
+        return await result_fut
+
+    def is_available(self):
+        return len([w for w in self.complete_request_workers if not w.done()]) > 0
\ No newline at end of file
