@@ -11,6 +11,7 @@ import time
 import asyncio
 
 from lmql.runtime.tokenizer import load_tokenizer
+from lmql.runtime.stats import Stats
 
 class OpenAIStreamError(Exception): pass
 class OpenAIRateLimitError(OpenAIStreamError): pass
@@ -22,6 +23,8 @@ Capacity.reserved = 0
 
 stream_semaphore = None
 
+api_stats = Stats("openai-api")
+
 class CapacitySemaphore:
     def __init__(self, capacity):
         self.capacity = capacity
@@ -210,7 +213,7 @@ async def chat_api(**kwargs):
                                 continue
                             if complete_chunk == "[DONE]": 
                                 return
-                            
+
                             n_chunks += 1
                             sum_chunk_times += time.time() - last_chunk_time
                             last_chunk_time = time.time()
@@ -268,7 +271,6 @@ async def chat_api(**kwargs):
 
                 if current_chunk.strip() == "[DONE]":
                     return
-                
                 try:
                     last_message = json.loads(current_chunk.strip())
                     message = last_message.get("error", {}).get("message", "")
@@ -337,9 +339,12 @@ async def completion_api(**kwargs):
 
                             if len(complete_chunk.strip()) == 0: 
                                 continue
-                            if complete_chunk == "[DONE]": 
+                            if complete_chunk == "[DONE]":
                                 return
                             
+                            if n_chunks == 0:
+                                api_stats.times["first-chunk-latency"] = api_stats.times.get("first-chunk-latency", 0) + (time.time() - stream_start)
+
                             n_chunks += 1
                             sum_chunk_times += time.time() - last_chunk_time
                             last_chunk_time = time.time()
