@@ -1,4 +1,4 @@
-from typing import List, Any, Union, Optional
+from typing import List, Any, Union, Optional, NamedTuple
 
 import asyncio
 import numpy as np
@@ -11,6 +11,11 @@ from .dclib_array import DataArray, Continuation, topk, alpha_length_normalized,
 
 detokenize_seqs = True
 
+@dataclass
+class DecoderGraphSnapshot:
+    node_hashes = None
+    step_not_updated_count = None
+
 class DecoderGraph:
     def __init__(self):
         self.nodes = {}
@@ -19,6 +24,8 @@ class DecoderGraph:
 
         self.ctr = 0
 
+        self.json_snapshot: DecoderGraphSnapshot = DecoderGraphSnapshot()
+
     def add_node(self, node):
         uid = f"n{self.ctr}"
         self.ctr += 1
@@ -34,13 +41,39 @@ class DecoderGraph:
     def add_edge(self, from_node, to_node):
         self.edges.append((self.node_ids[from_node], self.node_ids[to_node]))
 
-    async def json(self):
+    async def json(self, diff: bool = False):
         nodes = []
+        
+        if self.json_snapshot.node_hashes is None:
+            self.json_snapshot.node_hashes = {}
+        if self.json_snapshot.step_not_updated_count is None:
+            self.json_snapshot.step_not_updated_count = {}
+        
         for k, v in self.nodes.items():
+            hash = None
+            if diff and k in self.json_snapshot.node_hashes:
+                # this will ignore changes to nodes that have not been updated for 3 steps (may miss changes to nodes that 
+                # are not updated for a long time, careful for now)
+                if k in self.json_snapshot.step_not_updated_count and self.json_snapshot.step_not_updated_count[k] > 4:
+                    continue
+                hash = await v.json_hash()
+                if self.json_snapshot.node_hashes[k] == str(hash):
+                    self.json_snapshot.step_not_updated_count[k] = self.json_snapshot.step_not_updated_count.get(k, 0) + 1
+                    continue
+                    
+                # else:
+                #     print("node changed", k, self.json_snapshot.node_hashes[k], hash)
+            if hash is None:
+                hash = await v.json_hash()
+
             nodes.append({
                 "id": k,
                 **await v.json()
             })
+            
+            if diff:
+                self.json_snapshot.node_hashes[k] = str(hash)
+                self.json_snapshot.step_not_updated_count[k] = 0
 
         return {
             "nodes": nodes,
@@ -48,7 +81,7 @@ class DecoderGraph:
         }
 
 class DecoderSequence:
-    def __init__(self, input_ids_or_str, logprobs=None, deterministic=None, stop_phrase=None, predecessor=None, user_data=None, sticky_user_data_keys=None, epsilon_node=False):
+    def __init__(self, input_ids_or_str, logprobs=None, deterministic=None, stop_phrase=None, predecessor=None, user_data=None, sticky_user_data_keys=None, epsilon_node=False, internal=False):
         assert all([p > DecoderSequence.truncation_threshold for p in logprobs]) if logprobs is not None else True
 
         if type(input_ids_or_str) == str:
@@ -84,7 +117,7 @@ class DecoderSequence:
         # if no stop_phrase is provided assume no tokens to be stop_phrase
         if stop_phrase is None: self.stop_phrase = np.array([False for _ in range(len(input_ids))])
         else: self.stop_phrase = stop_phrase
-        assert len(self.stop_phrase) == len(self.input_ids), "Length of determinism status does not match length of inputs"
+        assert len(self.stop_phrase) == len(self.input_ids), "Length of stop_phrase status does not match length of inputs for {} and {}".format(self.stop_phrase, self.input_ids)
 
         # cache for logits when model is called with this sequence
         self.logits = None
@@ -95,7 +128,7 @@ class DecoderSequence:
 
         self.user_data = user_data
         self.sticky_user_data_keys = sticky_user_data_keys if sticky_user_data_keys is not None else set()
-        self.sticky_user_data_keys.add("head.variable")
+        # self.sticky_user_data_keys.add("head.variable")
 
         # alternative operation that may called to obtain a score in self.score_next_token if self.raw_logits are not available
         self.score_next_tokens_op = None
@@ -117,7 +150,7 @@ class DecoderSequence:
 
         setattr(s, "_id", f"s_{DecoderSequence.seq_ctr}")
 
-        if DecoderSequence.graph is not None:
+        if DecoderSequence.graph is not None and not kwargs.get("internal"):
             DecoderSequence.graph.add_node(s)
 
             predecessor = kwargs.get("predecessor")
@@ -180,6 +213,12 @@ class DecoderSequence:
         
         return self.tokenizer_cache[name]
 
+    async def json_hash(self):
+        o = await self.json()
+        if type(o["user_data"]) is dict and "openai-continuations" in o["user_data"]:
+            o["user_data"].pop("openai-continuations")
+        return hash(str(o))
+
     async def json(self):
         seqtext = await self.detokenized("seqtext")
         text = await self.detokenized("text")
@@ -233,6 +272,9 @@ class DecoderSequence:
                 except:
                     if hasattr(o, "json") and callable(o.json):
                         return await self.default(await o.json())
+                    # check for NamedTuple
+                    elif hasattr(o, "_asdict") and callable(o._asdict):
+                        return await self.default(o._asdict())
                     elif type(o) is int:
                         return o
                     elif type(o) is float:
@@ -249,7 +291,11 @@ class DecoderSequence:
         if continuation is None and user_data is None:
             user_data = {}
         elif continuation is not None:
-            user_data = deepcopy(continuation.user_data)
+            if type(continuation.user_data) is list:
+                assert len(continuation.user_data) == 1, f"continuation.user_data is a list of length {len(continuation.user_data)} but should be a list of length 1"
+                user_data = deepcopy(continuation.user_data[0])
+            else:
+                user_data = deepcopy(continuation.user_data)
         elif user_data is not None:
             user_data = deepcopy(user_data)
         else:
@@ -260,8 +306,9 @@ class DecoderSequence:
             set_path(user_data, sk, self.data(sk), create_missing=True, replace=False)
         return user_data
 
-    def extend(self, continuation):
+    def extend(self, continuation, internal=False):
         stop_phrase = self.detect_stop_phrase(continuation)
+
         return DecoderSequence(
             input_ids_or_str=np.concatenate([self.input_ids, continuation.token.reshape(1)]), 
             logprobs=np.concatenate([self.logprobs, continuation.logprob.reshape(1)]),
@@ -271,13 +318,15 @@ class DecoderSequence:
             stop_phrase=stop_phrase,
             predecessor=self,
             user_data=self.extend_user_data(continuation),
-            sticky_user_data_keys=self.sticky_user_data_keys
+            sticky_user_data_keys=self.sticky_user_data_keys,
+            internal=internal
         )
 
     def detect_stop_phrase(self, continuation):
         old_stop_phrase = self.stop_phrase
         new_stop_phrase = np.concatenate([old_stop_phrase, np.array([False])])
-        stop_phrases = self.data("head.stopping_phrases.tokenized")
+        
+        stop_phrases = self.data("head").stopping_phrases["tokenized"] if self.data("head") is not None else []
 
         if stop_phrases is None:
             return new_stop_phrase
@@ -298,8 +347,14 @@ class DecoderSequence:
         ids = ", ".join([str(i) for i in self.input_ids[-10:]])
         return f"<seq token_len={len(self.input_ids)} ids=[... {ids}]>"
 
-    async def text(self) -> str:
-        return str([await get_tokenizer().decode(self.input_ids)])
+    async def text(self, offset:int=None, limit:int=None, pretty=True) -> str:
+        offset = offset or 0
+        limit = limit or len(self.input_ids)
+        raw_text = await get_tokenizer().decode(self.input_ids[offset:limit])
+        if not pretty: 
+            return raw_text
+        else:
+            return str([raw_text])[2:-2]
 
     async def str(self, full=False) -> str:
         ids = ", ".join([str(i) for i in self.input_ids[-10:]])
@@ -336,6 +391,8 @@ def resolve_path(d, path):
         return None
     segments = path.split(".")
     for segment in segments:
+        if hasattr(d, "_asdict") and callable(d._asdict):
+            d = d._asdict()
         if segment not in d:
             return None
         d = d[segment]
@@ -376,40 +433,48 @@ def deepmerge(a, b):
     return a
 
 class DeterministicDecoderSequence(DecoderSequence):
-    def __init__(self, input_ids, logprobs, deterministic, stop_phrase, next_ids, next_logprobs=None, next_deterministic=None, predecessor=None, user_data=None, needs_rewrite=False, sticky_user_data_keys=None):
-        super().__init__(input_ids, logprobs, deterministic, stop_phrase, predecessor, user_data=user_data.copy() if user_data is not None else None, sticky_user_data_keys=sticky_user_data_keys)
+    def __init__(self, input_ids, logprobs, deterministic, stop_phrase, next_ids, next_logprobs=None, next_deterministic=None, predecessor=None, user_data=None, needs_rewrite=False, sticky_user_data_keys=None, internal=False):
+        super().__init__(input_ids, logprobs, deterministic, stop_phrase, predecessor, user_data=user_data.copy() if user_data is not None else None, sticky_user_data_keys=sticky_user_data_keys, internal=internal)
         self.next_ids = next_ids
         self.next_logprobs = next_logprobs
         self.next_deterministic = next_deterministic
 
         self.needs_rewrite = needs_rewrite
+        self.internal = internal
 
         if next_logprobs is not None: assert len(next_logprobs) == len(next_ids), "Length of deterministic continuation did not match length of provided logprobs"
         if next_deterministic is not None: assert len(next_deterministic) == len(next_ids), "Length of determinism status did not match length of provided logrprobs"
 
         self.align_user_data()
 
+    # async def text(self, offset: int = None, limit: int = None, pretty=True) -> str:
+    #     return "<detseq> " + await super().text(offset, limit, pretty)
+
     def align_user_data(self):
+        if self.user_data is None: return
+
         # lmql-specific user data should be different for deterministic sequences
         head_variable = resolve_path(self.user_data, "head.variable")
+
         if head_variable is not None:
-            if "before(" in head_variable:
-                head_variable = head_variable.split(":before(", 1)[0]
-            if head_variable == "__done__":
-                set_path(self.user_data, "head.variable", "__done__")
-            else:
-                if len(self.next_ids) > 0:
-                    set_path(self.user_data, "head.variable", head_variable + ":before(" + str(len(self.next_ids)) + ")")
-                else:
-                    set_path(self.user_data, "head.variable", head_variable)
+            # if "before(" in head_variable:
+            #     head_variable = head_variable.split(":before(", 1)[0]
+            # if head_variable == "__done__":
+            #     set_path(self.user_data, "head", self.user_data["head"].updated(variable="__done__"))
+            # else:
+            #     if len(self.next_ids) > 0:
+            #         set_path(self.user_data, "head", self.user_data["head"].updated(variable=head_variable + ":before(" + str(len(self.next_ids)) + ")"))
+            #     else:
+            #         set_path(self.user_data, "head", self.user_data["head"].updated(variable=head_variable))
             # deterministic sequences don't have stopping phrases
-            set_path(self.user_data, "head.stopping_phrases", {"tokenized": [], "text": []})
+            set_path(self.user_data, "head", self.user_data["head"].updated(stopping_phrases={"tokenized": [], "text": []}))
         else:
-            set_path(self.user_data, "head.variable", "<prompt>")
+            set_path(self.user_data, "head", self.user_data["head"].updated(variable="<prompt>"))
+        
         if len(self.next_ids) > 0:
-            set_path(self.user_data, "head.mask", "{token_id=" + str(self.next_ids[0]) + "}")
+            set_path(self.user_data, "head", self.user_data["head"].updated(mask="{token_id=" + str(self.next_ids[0]) + "}"))
         else:
-            set_path(self.user_data, "head.mask", "<not available yet>")
+            set_path(self.user_data, "head", self.user_data["head"].updated(mask="<not available yet>"))
     
     @property
     def is_query_constrained(self):
@@ -434,7 +499,7 @@ class DeterministicDecoderSequence(DecoderSequence):
             "logprob": self.logprobs[-1],
             "seqlogprob": self.logprobs.sum(),
             "pool": self.pool,
-            "deterministic": True,
+            **({"deterministic": True} if self.deterministic[-1] else {}),
             "next_ids": self.next_ids.tolist(),
             "next_logprobs": self.next_logprobs.tolist() if self.next_logprobs is not None else None,
             "user_data": await self.user_data_json(),
@@ -448,10 +513,11 @@ class DeterministicDecoderSequence(DecoderSequence):
             "sticky_user_data_keys": list(self.sticky_user_data_keys)
         }
 
-    def expand(self):
+    def expand(self, limit=None):
         """ Returns the sequence that corresponds to the full expansion of this pre-determined deterministic sequence. """
         s = self
-        while type(s) is DeterministicDecoderSequence and len(s.next_ids) > 0:
+        steps = 0
+        while type(s) is DeterministicDecoderSequence and len(s.next_ids) > 0 and (limit is None or steps < limit):
             s = s.extend(Continuation(s.next_ids[0], s.next_logprobs[0], s.user_data))
         return s
 
@@ -459,7 +525,7 @@ class DeterministicDecoderSequence(DecoderSequence):
         # additionally add self.user_data for deterministic sequence extensions
         return deepmerge(deepcopy(self.user_data), super().extend_user_data(continuation=continuation, user_data=user_data))
 
-    def extend(self, continuation):
+    def extend(self, continuation, internal=False):
         # if not more predetermined tokens are left, just extend as usual and return a regular seq()
         if len(self.next_ids) <= 0:
             return super().extend(continuation)
@@ -493,6 +559,7 @@ class DeterministicDecoderSequence(DecoderSequence):
             user_data=user_data,
             needs_rewrite=self.needs_rewrite,
             sticky_user_data_keys=self.sticky_user_data_keys,
+            internal=internal or self.internal
         )
 
     def __repr__(self) -> str:
@@ -567,7 +634,8 @@ def detseq(
     predecessor=None, 
     user_data=None,
     needs_rewrite=True,
-    sticky_user_data_keys=None):
+    sticky_user_data_keys=None,
+    internal=False):
     
     return DeterministicDecoderSequence(
         input_ids=ids, 
@@ -581,6 +649,7 @@ def detseq(
         user_data=user_data,
         needs_rewrite=needs_rewrite,
         sticky_user_data_keys=sticky_user_data_keys,
+        internal=internal
     )
 
 def seq(ids: List[int], logprobs:Optional[np.ndarray]=None, deterministic:Optional[np.ndarray]=None):
