@@ -1,14 +1,20 @@
+"""
+Runtime support used by compiled LMQL query code.
+"""
+
 import inspect
 from dataclasses import dataclass
 from typing import Any, Dict, Optional
 
 from lmql.ops.ops import *
-from lmql.runtime.output_writer import silent
-from lmql.runtime.prompt_interpreter import PromptInterpreter, LMQLResult
-from lmql.runtime.model_registry import LMQLModelRegistry
-from lmql.runtime.postprocessing.conditional_prob import ConditionalDistributionPostprocessor
-from lmql.runtime.postprocessing.group_by import GroupByPostprocessor
 from lmql.runtime.langchain import LMQLChainMixIn
+from lmql.runtime.model_registry import LMQLModelRegistry
+from lmql.runtime.output_writer import silent
+from lmql.runtime.interpreter import PromptInterpreter
+from lmql.runtime.postprocessing.conditional_prob import \
+    ConditionalDistributionPostprocessor
+from lmql.runtime.postprocessing.group_by import GroupByPostprocessor
+
 
 def register_model(identifier, ModelClass):
     LMQLModelRegistry.registry[identifier] = ModelClass
@@ -51,6 +57,8 @@ class LMQLQueryFunction(LMQLChainMixIn):
     function_context: Optional[FunctionContext] = None
 
     is_langchain_use: bool = False
+
+    lmql_code: str = None
     
     def __init__(self, fct, output_variables, postprocessors, scope, *args, **kwargs):
         # check for pydantic base class and do kw initialization then
@@ -92,15 +100,20 @@ class LMQLQueryFunction(LMQLChainMixIn):
             args_of_query = self.function_context.args_of_query
             scope = self.function_context.scope
 
+        # do not consider kwargs that are already set
+        argnames = [a for a in argnames if a not in kwargs.keys()]
+
         assert len(args) == len(argnames), f"@lmql.query {self.fct.__name__} expects {len(argnames)} positional arguments, but got {len(args)}."
         captured_variables = set(args_of_query)
         for name, value in zip(argnames, args):
             if name in args_of_query:
                 kwargs[name] = value
                 captured_variables.remove(name)
-    
+
+        # resolve remaining unset args from scope
         for v in captured_variables:
-            kwargs[v] = scope.resolve(v)
+            if not v in kwargs:
+                kwargs[v] = scope.resolve(v)
         
         if "output_writer" in kwargs:
             self.output_writer = kwargs["output_writer"]
@@ -120,8 +133,10 @@ class LMQLQueryFunction(LMQLChainMixIn):
         kwargs = self.make_kwargs(*args, **kwargs)
 
         interpreter = PromptInterpreter(force_model=self.model)
+        
         if self.output_writer is not None:
-            interpreter.output_writer = self.output_writer
+            kwargs["output_writer"] = self.output_writer
+        interpreter.set_extra_args(**kwargs)
 
         query_kwargs = {}
         for a in self.args:
@@ -142,13 +157,20 @@ class LMQLQueryFunction(LMQLChainMixIn):
                 results = await postprocessor.process(results, self.output_writer)
         
         interpreter.print_stats()
+        interpreter.dcmodel.save()
 
         return results
 
+def context_call(fct_name, *args, **kwargs):
+    return ("call:" + fct_name, args, kwargs)
+
+def interrupt_call(fct_name, *args, **kwargs):
+    return ("interrupt:" + fct_name, args, kwargs)
+
 def tag(t):
     return f"<lmql:{t}/>"
 
-def query(output_variables=None, group_by=None):
+def compiled_query(output_variables=None, group_by=None):
     if output_variables is None:
         output_variables = []
     
@@ -166,4 +188,5 @@ def query(output_variables=None, group_by=None):
                                  output_variables=output_variables, 
                                  postprocessors=postprocessors, 
                                  scope=LMQLInputVariableScope(fct, calling_frame))
-    return func_transformer
\ No newline at end of file
+    return func_transformer
+    
\ No newline at end of file
