@@ -1,23 +1,23 @@
 import asyncio
-import os
 import inspect
-import lmql.runtime.bopenai as openai
-from lmql.runtime.stats import Stats
-from dataclasses import dataclass
-from typing import Any, Callable, Optional, List, Union
-
+import os
 from collections import namedtuple
+from dataclasses import dataclass
+from typing import Any, Callable, List, Optional, Union
+
 import numpy as np
+
+import lmql.runtime.masks as masks
+import lmql.runtime.bopenai as openai
+import lmql.runtime.dclib as dc
+import lmql.utils.nputil as nputil
+from lmql.runtime.dclib.dclib_model import DcModel
+from lmql.runtime.dclib.dclib_seq import (DecoderSequence, deepcopy, deepmerge,
+                                          detseq, is_deterministic)
+from lmql.runtime.stats import Stats
+from lmql.runtime.tokenizer import load_tokenizer
 from lmql.utils import nputil
 
-from lmql.runtime.decoder_head import DecoderHead
-from lmql.runtime.rewriter import InputIdRewriter, ActivePromptingRewriter
-
-from lmql.runtime.tokenizer import load_tokenizer
-from lmql.runtime.dclib.lmql_adapter import QueryDcLibAdapter
-from lmql.runtime.dclib.dclib_model import DcModel
-from lmql.runtime.dclib.dclib_seq import deepcopy, deepmerge, is_deterministic, detseq, DecoderSequence
-import lmql.runtime.dclib as dc
 
 def is_allowed(m): 
     """
@@ -111,13 +111,12 @@ class DclibOpenAiModel(DcModel):
         
         self.model_identifier = "openai/" + self.model.model_identifier
 
-        self.model.chunk_size = kwargs.get("openai_chunksize", 32)
+        self.model.chunk_size = kwargs.get("openai_chunksize", 64)
+        self.model.nostop = kwargs.get("openai_nonstop", False)
         self.num_billed_tokens = {}
         self.num_requests = 0
 
         self.stats = Stats("openai")
-
-        # openai.Completion.set_chaos(0.05)
         openai.AsyncConfiguration.set_tokenizer(self.tokenize)
 
     def log_billable_tokens(self, n: int):
@@ -136,24 +135,24 @@ class DclibOpenAiModel(DcModel):
             - (token_id, "fixed"): Complete with the provided token_id with full probability 1.0 (no API use necessary)
 
         """
-        stopping_phrases = s.data("head.stopping_phrases.text")
+        stopping_phrases = s.data("head").stopping_phrases["text"]
 
         if mask is None:
             return CompletionCall("*", None, s.input_ids, kwargs, stopping_phrases=stopping_phrases)
 
         invert = False
-        num_allowed = is_allowed(mask).sum(axis=-1)
-        assert num_allowed > 0, "DclibOpenAiModel: encountered logits mask with no allowed tokens"
+        num_allowed = masks.mask_num_allowed(mask)
+        assert num_allowed > 0, "DclibOpenAiModel: encountered logits mask with no allowed tokens: mask: {} mask type:{}".format(mask, type(mask))
 
         if num_allowed == 1:
             # check for <eos> case
-            if is_allowed(mask[self.eos_token_id]):
+            if masks.mask_is_allowed(mask, self.eos_token_id):
                 return CompletionCall("fixed", self.eos_token_id, s.input_ids, kwargs, stopping_phrases=stopping_phrases)
             else:
                 # otherwise we can treat this as a score call
-                return CompletionCall("fixed", mask.argmax(axis=-1), s.input_ids, kwargs, stopping_phrases=stopping_phrases)
-        elif num_allowed < mask.shape[-1]:
-            if mask.shape[-1] - num_allowed > num_allowed:
+                return CompletionCall("fixed", masks.mask_get_only_allowed(mask), s.input_ids, kwargs, stopping_phrases=stopping_phrases)
+        elif num_allowed < self.tokenizer.vocab_size:
+            if self.tokenizer.vocab_size - num_allowed > num_allowed:
                 # if we have to mask more than half of the tokens, we should just invert the masking
                 invert = True
         else: # num_allowed == mask.shape[-1] (full vocabulary)
@@ -185,32 +184,49 @@ class DclibOpenAiModel(DcModel):
         self.score_queue.put_nowait((kwargs,result_fut))
         return await result_fut
 
-    async def _score_next_tokens(self, s, next_tokens):
+    async def _score_next_tokens(self, s, next_tokens, noscore=False):
+        if noscore:
+            return np.zeros(len(next_tokens), dtype=np.float32)
         return (await self.api_score(np.concatenate([s.input_ids, next_tokens], axis=0), len(s.input_ids)))
     
-    async def score(self, sqs: List[DecoderSequence], tokens: List[List[int]], max_batch_size=4, deterministic=False, stop_phrase=False, needs_rewrite=True):
+    async def score(self, sqs: List[DecoderSequence], tokens: List[List[int]], max_batch_size=4, deterministic: Union[bool, List[bool]]=False, stop_phrase=False, needs_rewrite=True, user_data=None, noscore=False, internal=False):
         assert len(sqs) == len(tokens), "Number of sequences and number of tokens to be scored must match, but got {} and {}".format(len(sqs), len(tokens))
-        
-        completion = [np.array(cont) for cont in tokens]
 
         def make_detseq(s, token_score, completion):
+            # compose deterministic flags
+            if type(deterministic) is bool:
+                deterministic_flags = np.concatenate([s.deterministic, np.array([deterministic])])
+                next_deterministic = np.array([deterministic] * len(completion[1:]))
+            else:
+                assert type(deterministic) is list and len(deterministic) == len(completion), "If deterministic is a list, it must have the same length as the number of tokens to be scored, but is {} and {}".format(deterministic, completion)
+                deterministic_flags = np.concatenate([s.deterministic, np.array(deterministic[:1])])
+                next_deterministic = np.array(deterministic[1:])
+
             return detseq(ids=np.concatenate([s.input_ids, completion[:1]], axis=0), 
                     next_ids=completion[1:],
                     logprobs=np.concatenate([s.logprobs, token_score[:1]], axis=0),
                     next_logprobs=token_score[1:],
-                    deterministic=np.concatenate([s.deterministic, np.array([deterministic])]),
-                    next_deterministic=np.array([deterministic] * len(completion[1:])),
+                    deterministic=deterministic_flags,
+                    next_deterministic=next_deterministic,
                     predecessor=s,
-                    user_data=None,
+                    user_data=user_data,
                     stop_phrase=np.concatenate([s.stop_phrase, np.array([stop_phrase])]),
                     needs_rewrite=needs_rewrite,
-                    sticky_user_data_keys=s.sticky_user_data_keys)
+                    sticky_user_data_keys=s.sticky_user_data_keys,
+                    internal=internal
+            )
         results = []
 
-        for s,compl,result in zip(sqs, completion, await asyncio.gather(*(self._score_next_tokens(s, compl) for s, compl in zip(sqs, completion)))):
-            results.append(make_detseq(s, result, compl))
+        async for (s, tokens, scores) in self.score_tokens(sqs, tokens, max_batch_size=max_batch_size, noscore=noscore):
+            results.append(make_detseq(s, scores, tokens))
 
         return results
+    
+    async def score_tokens(self, sqs: List[DecoderSequence], tokens: List[List[int]], max_batch_size=None, noscore=False):
+        completion = [np.array(cont) for cont in tokens]
+
+        for s, tokens,scores in zip(sqs, completion, await asyncio.gather(*(self._score_next_tokens(s, compl, noscore=noscore) for s, compl in zip(sqs, completion)))):
+            yield (s, tokens, scores)
 
     async def async_complete(self, completion_call: Union[CompletionCall, List[CompletionCall]], **kwargs) -> openai.response_buffer:
         assert type(completion_call) is CompletionCall
@@ -221,6 +237,7 @@ class DclibOpenAiModel(DcModel):
 
         temperature = completion_call.kwargs.get("temperature", 0.0)
         logprobs = completion_call.kwargs.get("logprobs", 5)
+        noscore = completion_call.kwargs.get("noscore", False)
 
         kwargs = {
             "model": self.model.model_identifier,
@@ -239,6 +256,9 @@ class DclibOpenAiModel(DcModel):
             pass
         elif mode == "complete": # complete with mask
             logit_bias = completion_call.api_mask
+            # reduce chunk size, if logit mask seems very sparse
+            if len(logit_bias) > 0 and max(logit_bias.values()) == 100 and len(logit_bias) < 10:
+                kwargs["max_tokens"] = min(kwargs["max_tokens"], 4)
             kwargs.update({"logit_bias": logit_bias})
         elif mode == "fixed": # complete with fixed token
             fixed_next_token = completion_call.logit_mask_or_fixed_id # special return value case for prepare function
@@ -247,21 +267,27 @@ class DclibOpenAiModel(DcModel):
                 return CompletionResult(openai.response_buffer.singleton(token=fixed_next_token, token_logprob=0), completion_call.continuation_type, completion_call.logit_mask_or_fixed_id)
             else:
                 fixed_next_token = nputil.ensure_array(fixed_next_token, dtype=np.int64)
-                logprob = (await self.api_score(np.concatenate([input_ids, fixed_next_token.reshape(1)], axis=0), len(input_ids)))
+                if noscore: logprob = 0.0
+                else: logprob = (await self.api_score(np.concatenate([input_ids, fixed_next_token.reshape(1)], axis=0), len(input_ids)))
                 return CompletionResult(openai.response_buffer.singleton(token=fixed_next_token, token_logprob=logprob), completion_call.continuation_type, completion_call.logit_mask_or_fixed_id)
-        elif len(completion_call.stopping_phrases) > 0:
+        else:
+            assert False, f"Internal openai API dispatcher returned an unknown completion mode {mode}"
+
+        if len(completion_call.stopping_phrases) > 0:
             if len(completion_call.stopping_phrases) > 4:
                 # same but blaming it more on OpenAI
                 print("warning: the number of stopping phrases that would need to be passed to the OpenAI API is greater than 4. Since the OpenAI API only supports up to 4 stopping phrases, the first 4 stopping phrases will be passed to the API. Other stopping phrases will also be enforced, but may lead to an increase in the number of tokens billed to the user.")
-            kwargs.update({"stop": completion_call.stopping_phrases[:4]})
-        else:
-            assert False, f"Internal openai API dispatcher returned an unknown completion mode {mode}"
+            # skip stopping phrases for more speculative execution
+            if not self.model.nostop:
+                kwargs.update({"stop": completion_call.stopping_phrases[:4]})
 
         # TODO: we are now overestimate the number of tokens billed to the user since we are not account for stopping phrases for the sake of streaming
         self.count_billed_tokens(input_ids.size + kwargs.get("max_tokens") * batch_size, self.model_identifier)
         
         if self.model_args.get("chatty_openai", False):
-            print("Completion with", kwargs)
+            args = kwargs.copy()
+            args["prompt"] = str([await self.detokenize(kwargs["prompt"])])[2:-2]
+            print(f"openai complete: {args}")
 
         return CompletionResult((await openai.async_buffer(await openai.Completion.create(**kwargs), tokenizer=self.tokenize_list))[input_ids.size:], completion_call.continuation_type, completion_call.logit_mask_or_fixed_id)
     
@@ -270,66 +296,8 @@ class DclibOpenAiModel(DcModel):
             return [[t[0]] for t in await self.model.tokenize(tokens)]
         return tokens
     
-    async def _complete(self, completion_call: Union[CompletionCall, List[CompletionCall]], **kwargs):
-        if type(completion_call) is list:
-            input_ids = np.stack([c.input_ids for c in completion_call], axis=0)
-            assert input_ids.ndim == 2, f"_complete expects input_ids to be a 1D tensor per completion call, when multiple completion calls are passed, got {input_ids.ndim}D tensor."
-            # all other call parameters are assumed to be the same
-            batch_size = len(completion_call)
-            completion_call = completion_call[0]
-        else:
-            batch_size = 1
-            input_ids = completion_call.input_ids
-            assert input_ids.ndim == 1, f"_complete expects input_ids to be a 1D tensor when only one completion_call is passed, got {input_ids.ndim}D tensor."
-
-        temperature = completion_call.kwargs.get("temperature", 0.0)
-        logprobs = completion_call.kwargs.get("logprobs", 5)
-
-        kwargs = {
-            "model": self.model.model_identifier,
-            "prompt": input_ids.tolist()[0], # no more batching at this point
-            "max_tokens": self.model.chunk_size,
-            "temperature": temperature,
-            "logprobs": logprobs,
-            "user": "lmql",
-            "stream": True,
-        }
-
-        mode = completion_call.mode
-        
-        if mode == "*": # complete without mask
-            pass
-        elif mode == "complete": # complete with mask
-            logit_bias = completion_call.api_mask
-            kwargs.update({"logit_bias": logit_bias})
-        elif mode == "fixed": # complete with fixed token
-            fixed_next_token = completion_call.logit_mask_or_fixed_id # special return value case for prepare function
-            # TODO revisit this, what kind of probability do we want here (masked or unmasked/scored)
-            if fixed_next_token == self.eos_token_id:
-                return OpenAIModelOutputBuffer.fixed_output(fixed_next_token, 0)
-            else:
-                if not nputil.is_array(fixed_next_token): 
-                    fixed_next_token = np.array(fixed_next_token)
-                logprob = (await self.api_score(np.concatenate([input_ids, fixed_next_token.reshape(1)], axis=0), len(input_ids))).result
-                return OpenAIModelOutputBuffer.fixed_output(fixed_next_token, logprob)
-        elif len(completion_call.stopping_phrases) > 0:
-            if len(completion_call.stopping_phrases) > 4:
-                # same but blaming it more on OpenAI
-                print("warning: the number of stopping phrases that would need to be passed to the OpenAI API is greater than 4. Since the OpenAI API only supports up to 4 stopping phrases, the first 4 stopping phrases will be passed to the API. Other stopping phrases will also be enforced, but may lead to an increase in the number of tokens billed to the user.")
-            kwargs.update({"stop": completion_call.stopping_phrases[:4]})
-        else:
-            assert False, f"Internal openai API dispatcher returned an unknown completion mode {mode}"
-
-        # TODO: we are now overestimate the number of tokens billed to the user since we are not account for stopping phrases for the sake of streaming
-        self.count_billed_tokens(input_ids.size + kwargs.get("max_tokens") * batch_size, self.model_identifier)
-        
-        if self.model_args.get("chatty_openai", False):
-            print("Completion with", kwargs)
-        
-        async def complete_op():
-            return openai_complete_create(**kwargs)
-        return await CompleteTask(complete_op, None, continuation_type=completion_call.continuation_type, 
-            logit_mask_or_fixed_id=completion_call.logit_mask_or_fixed_id).run(retries=3)
+    async def openai_cache_delegate(self, kwargs, tokens, scores):
+        print(tokens, scores, self.cache_delegate)
 
     def count_billed_tokens(self, n, model):
         if model not in self.num_billed_tokens.keys():
@@ -344,7 +312,7 @@ class DclibOpenAiModel(DcModel):
             with self.stats.timer("logit_masks"):
                 # print("completion_buffer", s)
                 constrained_seqs = np.array([s.is_query_constrained], dtype=np.bool_)
-                logits_mask_result = await self.compute_logits_mask(s.input_ids.reshape(1, -1), [s.user_data], constrained_seqs, **kwargs)
+                logits_mask_result = await self.compute_logits_mask(s.input_ids.reshape(1, -1), [s.user_data], constrained_seqs, [s], **kwargs)
                 logits_mask = logits_mask_result.logits_mask[0]
 
             # update user data with new information obtained when computing logits masks
@@ -377,16 +345,76 @@ class DclibOpenAiModel(DcModel):
                     None,
                 )
 
-            return await self.async_complete(completion_call)
+            completion_result = await self.async_complete(completion_call)
+            # eagerly expand and cache full completion if a cache_delegate is available
+            if self.cache_delegate is not None:
+                await self.expand_and_cache(s, completion_result, "top-1", logprobs=kwargs.get("logprobs", 1))
+            
+            assert not await completion_result.buffer.empty(), "Completion result is empty on arrival"
+            return completion_result
 
         return await asyncio.gather(*[get_buffer(i, s) for i, s in enumerate(seqs)])
 
+    async def expand_and_cache(self, s: DecoderSequence, completion_result: CompletionResult, edge_type, logprobs=1):
+        async def token_stream():
+            nonlocal edge_type, s, completion_result
+            response_buffer = completion_result.buffer
+            
+            try:
+                tokens = []
+                scores = []
+
+                while True:
+                    try:
+                        if await response_buffer.empty():
+                            break
+
+                        res = await response_buffer.get(0)
+                        tokens = nputil.ensure_iterable(res["logprobs"]["tokens"])
+                        scores = res["logprobs"]["token_logprobs"]
+
+                        topprobs = res["logprobs"]["top_logprobs"]
+                        if topprobs is not None and logprobs > 1:
+                            topk_tokens = list(topprobs.items())
+                            topk_tokens = [(tok[0], score) for (tok_str, score), tok in zip(topk_tokens, await self.tokenize_list([s for s,_ in topk_tokens]))]
+                            topk_tokens += [(tokens[0], scores)]
+                            topk_tokens = list(dict.fromkeys(topk_tokens))
+                            topk_tokens = sorted(topk_tokens, key=lambda x: x[1], reverse=True)
+                            topk_tokens = topk_tokens[:logprobs]
+                            
+                            tokens = [tok for tok, _ in topk_tokens]
+                            scores = [score for _, score in topk_tokens]
+                            edge_type = ["top-{}".format(i+1) for i in range(len(topk_tokens))]
+                        
+                        # future continuation
+                        response_buffer = response_buffer[1:]
+                        continuation = CompletionResult(response_buffer, completion_result.continuation_type, completion_result.logit_mask_or_fixed_id)
+
+                        if continuation.continuation_type is None:
+                            edge_type = None
+
+                        user_data = {
+                            "openai-continuations": {
+                                continuation.continuation_type: continuation
+                            }
+                        }
+                        yield (s, tokens, scores, edge_type, user_data)
+                    except IndexError:
+                        break
+                # print("fully expanded speculativate continuation:", [await self.detokenize(tokens)], flush=True)
+                # if len(tokens) > 1:
+                #     await self.cache(s, tokens, scores, edge_type)
+            except Exception as e:
+                import traceback
+                traceback.print_exc()
+                print("error in expand_and_cache worker:", e, flush=True)
+        self.register_token_stream(token_stream)
+
     async def argmax(self, sequences, **kwargs):
-        return await self.sample(sequences, num_samples=1, temperature=0)
+        return await self.sample(sequences, num_samples=1, temperature=0, **kwargs)
 
     def report_stats(self, printer, decoder_step=None):
         if printer is None:
-            print("no printer")
             return
         if hasattr(printer, "report_model_stats"):
             s = openai.Completion.get_stats()
@@ -407,7 +435,7 @@ class DclibOpenAiModel(DcModel):
         kwargs = {**self.model_args, **kwargs}
 
         async def op_sample(seqs):
-            completions: List[CompletionResult] = await self.completion_buffer(seqs, logprobs=5, **kwargs)
+            completions: List[CompletionResult] = await self.completion_buffer(seqs, logprobs=1, **kwargs)
             
             next_token_ids = []
             next_token_scores = []
@@ -429,7 +457,7 @@ class DclibOpenAiModel(DcModel):
                     next_token_ids.append(np.array([next_token], dtype=np.int64))
                     next_token_scores.append(np.array([next_token_score], dtype=np.float32))
                     
-                    full_logits = np.ones(self.model.tokenizer.vocab_size) * np.finfo(np.float32).min
+                    full_logits = np.ones(self.model.get_tokenizer().vocab_size) * np.finfo(np.float32).min
                     if next_token < len(full_logits):
                         full_logits[next_token] = next_token_score
                     # else: 
@@ -447,14 +475,14 @@ class DclibOpenAiModel(DcModel):
                 prob_tokens = await self.tokenize_list(tokens)
                 token_ids = np.array(prob_tokens, dtype=np.int64).reshape(-1)
 
-                full_logits = np.ones(self.model.tokenizer.vocab_size) * np.finfo(np.float32).min
+                full_logits = np.ones(self.model.get_tokenizer().vocab_size) * np.finfo(np.float32).min
                 full_logits[token_ids] = np.array(logprobs)
                 full_logits[next_token] = np.finfo(np.float32).min
                 assert kwargs.get("temperature", 1.0) != 0.0 or np.all(full_logits < next_token_score), "next token score is not the highest"
 
                 # retroactively apply logits mask to logits
                 mask = completion.logit_mask_or_fixed_id
-                if mask is None: 
+                if mask is None:
                     pass
                 elif type(mask) is int: 
                     full_logits[mask] = np.finfo(np.float32).min
@@ -477,7 +505,7 @@ class DclibOpenAiModel(DcModel):
             next_token_scores = next_token_scores
 
             def successor_user_data(continuation_buffer: SequenceResult, num_successors):
-                default_user_data = {"head": {"variable": s.data("head.variable"), "head_index": s.data("head.head_index")}}
+                default_user_data = {}
                 if continuation_buffer.continuation_type is None:
                     return [default_user_data.copy()] * num_successors
                 continuation_as_user_data = {
@@ -488,9 +516,8 @@ class DclibOpenAiModel(DcModel):
                 }
                 return [continuation_as_user_data] + [default_user_data.copy()] * (num_successors - 1)
 
-            s = [s.make_successors(next_token_ids[i], next_token_scores[i], logits=logits[i], 
+            return [s.make_successors(next_token_ids[i], next_token_scores[i], logits=logits[i], 
                 user_data=successor_user_data(continuation_buffers[i], len(next_token_ids[i]))) for i,s in enumerate(seqs)]
-            return s
         with self.stats.timer("sample"):
             return await sequences.aelement_wise(op_sample)
 
@@ -527,7 +554,7 @@ class DclibOpenAiModel(DcModel):
                     next_token_ids.append(np.array([next_token], dtype=np.int64))
                     next_token_scores.append(np.array([next_token_score], dtype=np.float32))
                     
-                    full_logits = np.ones(self.model.tokenizer.vocab_size) * np.finfo(np.float32).min
+                    full_logits = np.ones(self.model.get_tokenizer().vocab_size) * np.finfo(np.float32).min
                     full_logits[next_token] = next_token_score
                     logits.append(full_logits)
                     continue
@@ -539,11 +566,11 @@ class DclibOpenAiModel(DcModel):
                 probs = sorted(list(complete_data["logprobs"]["top_logprobs"].items()), key=lambda x: x[1], reverse=True)
                 logprobs = [p[1] for p in probs]
                 tokens = [p[0] for p in probs]
-                token_ids = np.array([self.model.tokenizer(t)["input_ids"][0] for t in tokens], dtype=np.int64)
+                token_ids = np.array([self.model.get_tokenizer()(t)["input_ids"][0] for t in tokens], dtype=np.int64)
 
                 assert token_ids[0] == next_token, f"top1 logprob token is not the same as the predicted token {token_ids[0]} (top1) != {next_token} (predicted)"
 
-                full_logits = np.ones(self.model.tokenizer.vocab_size) * np.finfo(np.float32).min
+                full_logits = np.ones(self.model.get_tokenizer().vocab_size) * np.finfo(np.float32).min
                 full_logits[token_ids] = np.array(logprobs)
 
                 # retroactively apply logits mask to logits
@@ -552,18 +579,18 @@ class DclibOpenAiModel(DcModel):
                 elif type(mask) is int: full_logits[mask] = np.finfo(np.float32).min
                 else: full_logits[mask < 0] = np.finfo(np.float32).min
                 
+                # make sure all token_ids are unique
+                token_ids = np.array(list(set(token_ids)))
+
                 # re-determine logprobs with logits mask applied
                 logprobs = np.take_along_axis(full_logits, token_ids, axis=-1)
 
-                # make sure all token_ids are different
-                assert len(token_ids) == len(set(token_ids)), "token_ids are not unique"
-
                 next_token_ids.append(token_ids)
                 next_token_scores.append(logprobs)
                 logits.append(full_logits)
 
             def successor_user_data(continuation_buffer: SequenceResult, num_successors):
-                default_user_data = {"head": {"variable": s.data("head.variable"), "head_index": s.data("head.head_index")}}
+                default_user_data = {}
                 if continuation_buffer.continuation_type is None:
                     return [default_user_data.copy()] * num_successors
                 continuation_as_user_data = {
@@ -798,6 +825,7 @@ class OptimisticChunkBasedOpenAIModel:
     def __init__(self, model_identifier, tokenizer):
         self.model_identifier = model_identifier.split("openai/",1)[1]
         self.chunk_size = 32
+        self.nostop = False
         self.tokenizer = tokenizer
 
         self.input_ids = []
@@ -812,7 +840,6 @@ class OptimisticChunkBasedOpenAIModel:
         def task():
             return self.tokenizer.encode(*args, **kwargs)
         return await asyncio.get_event_loop().run_in_executor(None, task)
-        return self.tokenizer.tokenize(*args, **kwargs)
     
     async def detokenize(self, *args, **kwargs):
         def task():
@@ -841,39 +868,6 @@ class OptimisticChunkBasedOpenAIModel:
         complete_op = lambda: openai_complete_create(**kwargs)
         return CompleteTask(complete_op, None, continuation_type=None).run(retries=3)
         
-
-    async def advance(self):
-        sequence_indices_to_extend_on = []
-        prompts_to_extend = []
-        for i, buffer in enumerate(self.buffer):
-            if await buffer.at_end():
-                sequence_indices_to_extend_on.append(i)
-                prompts_to_extend.append(self.input_ids[i])
-
-        if len(prompts_to_extend) > 0:
-            res = self._complete(np.stack(prompts_to_extend, axis=0))
-            response_buffer = OpenAIModelOutputBuffer(res, len(prompts_to_extend), self.tokenizer)
-            for i, seq_idx in enumerate(sequence_indices_to_extend_on):
-                self.buffer[seq_idx] = response_buffer.buffer(i)
-
-        # advance all self.input_ids by one token from buffer
-        next_tokens = []
-        for i, buffer in enumerate(self.buffer):
-            assert not await buffer.at_end(), f"openai model did not complete decoder head {i} further."
-            token, logprob = await buffer.pop()
-            next_tokens.append(token)
-            self.next_token_scores[i].append(logprob)
-
-        next_tokens = np.array(next_tokens, dtype=np.int64)
-        self.input_ids = np.concatenate([self.input_ids, next_tokens.unsqueeze(1)], axis=1)
-
-    def make_api_logits_mask(self, logits, invert):
-        if invert:
-            masked = (logits >= 0)
-        else:
-            masked = (logits < 0)
-        mask_value = 100 if invert else -100
-        return {int(idx): mask_value for idx in np.nonzero(masked)[0]}
     
     async def tokenize(self, text):
         return self.tokenizer(text)["input_ids"]
@@ -919,59 +913,29 @@ def openai_model(model_identifier):
     # make sure openai org and secret are available
     import lmql.runtime.openai_secret
     
-    class ServedModelInterfaceCls:
+    class OpenAIModel:
         def __init__(self) -> None:
             self.model_identifier = model_identifier
-            self.decoder_args = None
-
-            self.use_dclib_decoders = False
-            self.active_prompting = False
-
             self.served_model = None
             self._tokenizer = None
 
-            self.bos_token_id = self.get_tokenizer().bos_token_id
-            self.eos_token_id = self.get_tokenizer().eos_token_id
-
-            self.adapter = QueryDcLibAdapter(self.get_tokenizer().vocab_size, self.tokenize, self.detokenize, self.bos_token_id, self.eos_token_id)
-            self.hf_stats = HFModelStatsAdapter()
+            self.decoder_args = {
+                "openai_chunksize": 64
+            }
 
         def get_tokenizer(self):
             if self._tokenizer is None:
                 self._tokenizer = load_tokenizer("gpt2")
-            self.served_model = OptimisticChunkBasedOpenAIModel(model_identifier, self._tokenizer)
+            self.served_model = self
             return self._tokenizer
 
-        def set_decoder(self, method, **kwargs):
-            # defaults
-            DEFAULTS = {
-                "max_len": 512,
-                "decoder": method
-            }
-            self.decoder_args = DEFAULTS.copy()
-
-            # custom user arguments
-            protected_args = set(["input_ids", "additional_logits_processors", "bos_token_id", "eos_token_id"])
-            for key, value in kwargs.items():
-                if key in protected_args:
-                    print("warning: cannot override runtime determined decoder argument {}.".format(key))
-                    continue
-                self.decoder_args[key] = value
-        
         def get_dclib_model(self):
             bos_token_id = self.get_tokenizer().bos_token_id
             eos_token_id = self.get_tokenizer().eos_token_id
 
-            return DclibOpenAiModel(self.served_model, bos_token_id, eos_token_id, **self.decoder_args)
+            dc.set_dclib_tokenizer(dc.tokenizer("lmql-adapter-tokenizer", self.tokenize, self.detokenize, bos_token_id, eos_token_id))
 
-        async def score_distribution_values(self, prompt, values):
-            return await self.adapter.score_distribution_values(prompt, values, self.get_dclib_model())
-
-        async def query(self, prompt, mask_logits_processor, head_input_id_rewriter, active_prompt_rewriter):
-            assert self.decoder_args is not None, "Cannot query() a model without calling set_decoder first."
-
-            dclib_model = self.get_dclib_model()
-            return await self.adapter.query(prompt, mask_logits_processor, head_input_id_rewriter, active_prompt_rewriter, dclib_model, self.decoder_args)
+            return DclibOpenAiModel(self, self.get_tokenizer(), **self.decoder_args)
 
         async def tokenize(self, text):
             return self.get_tokenizer()(text)["input_ids"]
@@ -981,4 +945,4 @@ def openai_model(model_identifier):
 
         def sync_tokenize(self, text):
             return self.get_tokenizer()(text)["input_ids"]
-    return ServedModelInterfaceCls
\ No newline at end of file
+    return OpenAIModel
\ No newline at end of file
