@@ -1,58 +1,69 @@
-def get_js_tokenizer(model_identifier):
-    import js
-    from pyodide.ffi import to_js
+import asyncio
+from lmql.runtime.caching import cache_file_exists, cachefile
 
-    assert "gpt" in model_identifier, "JS tokenizer only supports GPT models."
+class PythonBackedTokenizer:
+    """ Custom tokenizer to be used only in a browser environment. This tokenizer only supports GPT tokenization. """
+    def __init__(self, model_identifier):
+        import gpt3_tokenizer
+        assert "gpt" in model_identifier, "PythonBackedTokenizer only supports GPT family models"
 
-    class JSBridgedTokenizer:
-        """ Custom tokenizer to be used only in a browser environment. This tokenizer only supports GPT tokenization. """
-        def __init__(self):
-            self.bos_token_id = 50256
-            self.eos_token_id = 50256
-            self._vocab = None
+        self.bos_token_id = 50256
+        self.eos_token_id = 50256
+        self._vocab = None
 
-        @property
-        def vocab_size(self):
-            return len(self.vocab)
+    @property
+    def vocab_size(self):
+        return len(self.vocab)
 
-        @property
-        def vocab(self):
-            if self._vocab is None:
-                self._vocab = js.get_vocab().to_py()
-            return self._vocab
+    @property
+    def vocab(self):
+        if self._vocab is None:
+            import gpt3_tokenizer
+            self._vocab = gpt3_tokenizer._entry._encoder
+        return self._vocab
 
-        def convert_tokens_to_string(self, tokens):
-            return js.convert_tokens_to_string_gpt(to_js(tokens))
+    def convert_tokens_to_string(self, tokens):
+        import gpt3_tokenizer
+        text_with_bytes = "".join(tokens)
+        textarr = [int(gpt3_tokenizer._entry._byte_decoder[x]) for x in list(text_with_bytes)]
+        text = bytearray(textarr).decode("utf-8")
+        return text
 
-        def tokenize(self, s):
-            unpack = False
-            if type(s) is not list:
-                s = [s]
-                unpack = True
-            tokens = [js.tokenize_gpt_toks(se).to_py() for se in s]
-            
-            if unpack:
-                return tokens[0]
-            else:
-                return tokens
-         
-        def decode(self, input_ids):
-            # set typed array type of input_ids to int
-            return str(js.detokenize_gpt(to_js([int(i) for i in input_ids])))
+    def tokenize(self, s):
+        if ("<|endoftext|>" in s):
+            return [50256];
+        import gpt3_tokenizer
+        unpack = False
+        if type(s) is not list:
+            s = [s]
+            unpack = True
 
-        def __call__(self, s: str):
-            unpack = False
-            if type(s) is not list:
-                s = [s]
-                unpack = True
-            input_ids = [[int(v) for v in js.tokenize_gpt(se)] for se in s]
-            
-            if unpack:
-                return {"input_ids": input_ids[0]}
-            else:
-                return {"input_ids": input_ids}
-    
-    return JSBridgedTokenizer()
+        tokens = [[gpt3_tokenizer._entry._decoder[i] for i in gpt3_tokenizer.encode(se)] for se in s]
+
+        if unpack:
+            return tokens[0]
+        else:
+            return tokens
+        
+    def decode(self, input_ids, clean_up_tokenization_spaces=None):
+        import gpt3_tokenizer
+        return gpt3_tokenizer.decode(input_ids)
+
+    def __call__(self, s: str, add_special_tokens=False):
+        import gpt3_tokenizer
+        
+
+        unpack = False
+        if type(s) is not list:
+            s = [s]
+            unpack = True
+        
+        input_ids = [gpt3_tokenizer.encode(se) if se != "<|endoftext|>" else [self.eos_token_id] for se in s]
+        
+        if unpack:
+            return {"input_ids": input_ids[0]}
+        else:
+            return {"input_ids": input_ids}
 
 global special_token_mappings
 special_token_mappings = {}
@@ -60,12 +71,21 @@ global reverse_special_token_mappings
 reverse_special_token_mappings = {}
 
 class LMQLTokenizer:
-    def __init__(self, tokenizer_impl):
+    INVALID_CHARACTER = "\uFFFD"
+
+    def __init__(self, tokenizer_impl, model_identifier):
         self.tokenizer_impl = tokenizer_impl
+        self.model_identifier = model_identifier
+        self.detokenizer_cache = {}
+
+        self._vocab = get_vocab(self.tokenizer_impl)
+        self.vocab_range = max(self._vocab.values()) + 1
 
     @property
     def vocab_size(self):
-        return self.tokenizer_impl.vocab_size
+        # in LMQL vocab_size is the vocab_range (the highest vocabulary ID + 1)
+        # this allows us to use a dense one hot array where no IDs are skipped
+        return self.vocab_range
 
     @property
     def bos_token_id(self):
@@ -92,28 +112,56 @@ class LMQLTokenizer:
         return tokens
         
     def decode(self, input_ids):
+        key = str(input_ids)
+        n = len(input_ids)
+        if n in self.detokenizer_cache.keys():
+            if key in self.detokenizer_cache[n].keys():
+                # print("cache hit")
+                return self.detokenizer_cache[n][key]
+        if n-1 in self.detokenizer_cache.keys():
+            key = str(input_ids[:-1])
+            if key in self.detokenizer_cache[n-1].keys():
+                global reverse_special_token_mappings
+                # print("secondary cache hit")
+                if input_ids[-1] >= self.vocab_size:
+                    extended = self.detokenizer_cache[n-1][key] + "<" + reverse_special_token_mappings[input_ids[-1]] + "/>"
+                else:
+                    extended = self.detokenizer_cache[n-1][key] + self.tokenizer_impl.decode([input_ids[-1]], clean_up_tokenization_spaces=False)
+                    if self.INVALID_CHARACTER in extended:
+                        return self.detokenizer_cache[n-1][key]
+                if not n in self.detokenizer_cache.keys():
+                    self.detokenizer_cache[n] = {}
+                self.detokenizer_cache[n][str(input_ids)] = extended
+                return extended
+
         s = ""
         for chunk in self.chunk_out_by_special_ids(input_ids):
             if type(chunk) is str:
                 s += chunk
             else:
-                s += self.tokenizer_impl.decode(chunk)
+                s += self.tokenizer_impl.decode(chunk, clean_up_tokenization_spaces=False)
+
+        if not n in self.detokenizer_cache.keys():
+            self.detokenizer_cache[n] = {}
+        self.detokenizer_cache[n][key] = s
+
         return s
 
-    def __call__(self, s: str):
+    def __call__(self, s: str, add_special_tokens=False):
         input_ids = []
         unpack = False
         if type(s) is not list:
             s = [s]
             unpack = True
-
+        
         for seq in s:
             chunk_input_ids = []
             for chunk in self.chunk_out_by_tags(seq):
                 if type(chunk) is int:
                     chunk_input_ids.append(chunk)
                 else:
-                    chunk_input_ids += self.tokenizer_impl(chunk)["input_ids"]
+                    result = self.tokenizer_impl(chunk, add_special_tokens=add_special_tokens)["input_ids"]
+                    chunk_input_ids += result
             input_ids.append(chunk_input_ids)
         
         if unpack:
@@ -128,7 +176,7 @@ class LMQLTokenizer:
         if identifier not in special_token_mappings:
             if len(special_token_mappings) == 0:
                 # offset vocabulary IDs by at least the next decimal power of 10
-                offset = 10 ** (len(str(self.vocab_size)))
+                offset = 10 ** (len(str(self.vocab_range)))
                 special_token_mappings[identifier] = offset
                 reverse_special_token_mappings[offset] = identifier
             else:
@@ -165,34 +213,47 @@ class LMQLTokenizer:
         segments.append(s[offset:])
         return segments
 
+def load_tokenizer_notransformers(model_identifier):
+    return PythonBackedTokenizer(model_identifier)
+
 def load_tokenizer(model_identifier):
     import os
 
-    # check environment of USE_JS_TOKENIZER
-    if "LMQL_BROWSER" in os.environ:
-        return LMQLTokenizer(get_js_tokenizer(model_identifier))
-
-    from transformers import AutoTokenizer
-    import torch
-
     # first try to load pickled tokenizer from cache (faster)
     import pickle
     import pathlib
 
-    cache_dir = pathlib.Path.home() / ".cache" / "lmql"
-    cache_dir.mkdir(parents=True, exist_ok=True)
     cache_identifier = model_identifier.replace("/", "-")
-    cache_path = cache_dir / f"tokenizer-{cache_identifier}.pkl"
+    cache_path = f"tokenizer-{cache_identifier}.pkl"
 
-    if cache_path.exists():
-        with open(cache_path, "rb") as f:
-            return LMQLTokenizer(pickle.load(f))
-    else:
+    try:
+        import torch
         from transformers import AutoTokenizer
-        t = AutoTokenizer.from_pretrained(model_identifier)
-        with open(cache_path, "wb") as f:
-            pickle.dump(t, f)
-        return LMQLTokenizer(t)
+
+        if cache_file_exists(cache_path):
+            with cachefile(cache_path, "rb") as f:
+                return LMQLTokenizer(pickle.load(f), model_identifier)
+        else:
+            t = AutoTokenizer.from_pretrained(model_identifier)
+
+            with cachefile(cache_path, "wb") as f:
+                pickle.dump(t, f)
+    except Exception as e:
+        # print("info: trying to use python-based tokenizer as 'transformers' is not available")
+        # fallback to non-transformers tokenizer
+        t = load_tokenizer_notransformers(model_identifier)
+
+    return LMQLTokenizer(t, model_identifier)
+
+def get_vocab(tokenizer):
+    if hasattr(tokenizer, "vocab"):
+        return tokenizer.vocab
+    elif hasattr(tokenizer, "get_vocab"):
+        return tokenizer.get_vocab()
+    elif hasattr(tokenizer, "tokenizer_impl"):
+        return get_vocab(tokenizer.tokenizer_impl)
+    else:
+        assert False, "Could not obtain full vocabulary from unknown tokenizer type: {}".format(type(tokenizer))
 
 if __name__ == "__main__":
     import sys
@@ -220,4 +281,4 @@ if __name__ == "__main__":
                 n += 1
                 result += f""""{t}","""
         print(n)
-        print(result)
\ No newline at end of file
+        print(result)
