@@ -4,11 +4,12 @@ import styled from 'styled-components';
 import Editor from "@monaco-editor/react";
 import React, { useEffect, useRef, useState } from "react";
 import { registerLmqlLanguage } from "./editor/lmql-monaco-language";
-import { BsSquare, BsFillExclamationSquareFill, BsBoxArrowUpRight, BsArrowRightCircle, BsCheckSquare, BsSendFill, BsFileArrowDownFill, BsKeyFill, BsTerminal, BsFileCode, BsGithub, BsCardList, BsFullscreen, BsXCircle, BsFillChatLeftTextFill, BsGear, BsGridFill } from 'react-icons/bs';
+import { BsSquare, BsFillExclamationSquareFill, BsBoxArrowUpRight, BsArrowRightCircle, BsFillCameraFill, BsCheckSquare, BsSendFill, BsFileArrowDownFill, BsKeyFill, BsTerminal, BsFileCode, BsGithub, BsCardList, BsFullscreen, BsXCircle, BsFillChatLeftTextFill, BsGear, BsGridFill } from 'react-icons/bs';
 import { DecoderGraph } from './DecoderGraph';
 import { BUILD_INFO } from './build_info';
 import exploreIcon from "./explore.svg"
 import { ExploreState, Explore, PromptPopup, Dialog } from './Explore'
+import { CodeScreenshot } from "./CodeScreenshot";
 import { errorState, persistedState, trackingState, displayState} from "./State"
 import { configuration, LMQLProcess, isLocalMode} from './Configuration';
 import { ValidationGraph } from "./ValidationGraph";
@@ -41,13 +42,12 @@ const ContentContainer = styled.div`
   
   color: white;
   flex: 1;
-  height: calc(100% - 40pt);
-  width: 100%;
   position: fixed;
   top: 0;
   left: 0;
-  width: 100%;
-  height: 100%;
+  width: calc(100% - 4pt);
+  height: calc(100% - 2pt);
+  margin: 2pt;
 `;
 
 const Panel = styled.div.attrs(props => ({ className: "panel" }))`
@@ -57,7 +57,8 @@ const Panel = styled.div.attrs(props => ({ className: "panel" }))`
   padding: 10px;
   display: flex;
   flex-direction: column;
-  margin-left: 2.5pt;
+  margin-left: 1.5pt;
+  margin-right: 1.5pt;
   width: 40%;
   position: relative;
 
@@ -169,16 +170,17 @@ const TokenCountDiv = styled.div`
   white-space: pre-line;
   max-height: 10pt;
 
-  :hover .tooltip {
-    visibility: visible;
-    opacity: 1;
-  }
+  /* indicate copy */
+  cursor: pointer;
 `
 
 function TokenCountIndicator() {
   const [stats, setStats] = useState({})
 
   const format_cost = (c, precision) => {
+    if (c == 0) {
+      return "$0.00"
+    }
     c = c.toFixed(precision)
     if (c === (0).toFixed(precision))
       return "<$" + (Math.pow(10, -precision)).toFixed(precision);
@@ -261,7 +263,8 @@ function TokenCountIndicator() {
   let tokenCount = 0;
   let model = ""
   let steps = 1;
-  if (stats.tokens) {
+  let copyString = ""
+  if (stats.tokens || stats._step) {
     tokenCount = stats.tokens
     model = stats.model
     steps = stats._step || 1
@@ -273,20 +276,26 @@ function TokenCountIndicator() {
     const toFirstUpper = k => k.charAt(0).toUpperCase() + k.slice(1)
     text = `Tokens: ${tokenCount}, ${otherKeys.map(k => `${toFirstUpper(k)}: ${stats[k]}`).join(", ")}`
 
+    copyString = `Tokens: ${tokenCount}\n${otherKeys.map(k => `${toFirstUpper(k)}: ${stats[k]}`).join("\n")}`
+
     // time elapsed
     if (stats._start) {
       const end = stats._end || stats._now || Date.now();
       const elapsed = (end - stats._start) / 1000
       text += `\n Time: ${elapsed.toFixed(1)}s, `
+      copyString += `\nTime: ${elapsed.toFixed(1)}s, `
     }
 
     text += `${(tokenCount / steps).toFixed(2)} tok/step`
     if (model.includes("openai")) {
       text += ` Est. Cost ${cost_estimate(model, tokenCount / 1000, 4)}`
+      copyString += `\nEst. Cost ${cost_estimate(model, tokenCount / 1000, 4)}`
     }
   }
 
-  return <TokenCountDiv>
+  return <TokenCountDiv onClick={() => {
+    navigator.clipboard.writeText(copyString)
+  }}>
     {text}
   </TokenCountDiv>
 }
@@ -400,7 +409,7 @@ function EditorPanel(props) {
 
   function handleEditorDidMount(editor, monaco) {
     ResizeObservers.addResizeListener(() => {
-      let fontSize = window.innerWidth < 700 ? 10 : 16
+      let fontSize = window.innerWidth < 700 ? 12 : 16
       editor.updateOptions({ "fontSize": fontSize })
       window.setTimeout(() => {
         editor.layout()
@@ -452,6 +461,8 @@ function EditorPanel(props) {
           // tabSize: 6,
           // show whitespace
           renderWhitespace: "all",
+          // set font family
+          fontFamily: "Fira Code, monospace",
           automaticLayout: true
         }}
         // custom language
@@ -549,6 +560,7 @@ const Row = styled.div`
       flex-direction: column;
       height: calc(100%);
       padding: 0;
+      margin-right: 4pt;
     }
 
     &.simple-mode.simple .panel {
@@ -681,6 +693,7 @@ const ModelResultText = styled.div`
   line-height: 1.6em;
   white-space: pre-wrap;
   overflow-y: auto;
+  font-size: 10pt;
 
   &.chat-mode {
     padding-bottom: 50pt;
@@ -762,9 +775,9 @@ const ModelResultText = styled.div`
 
   div .tag-assistant {
     display: inline-block;
-    width: 65%;
     border: 1pt solid #5c5c5c;
     margin-top: 5pt;
+    margin-right: 4%;
 
     border-radius: 8pt;
     overflow: hidden;
@@ -773,7 +786,7 @@ const ModelResultText = styled.div`
 
   div .tag-user {
     display: block;
-    margin-left: 32%;
+    margin-left: 4%;
     position: relative;
     border: 1pt solid #5c5c5c;
     border-radius: 8pt;
@@ -796,20 +809,16 @@ const ModelResultText = styled.div`
     background-color: #333;
     opacity: 1.0;
     border-radius: 2pt;
-    margin-left: 1pt;
-  }
-
-  div .variable:hover {
-    position: relative;
+    margin-left: 2pt;
   }
 
   div .badge {
-    padding: 2.5pt 4pt;
+    padding: 1.0pt 4pt;
     border-radius: 2pt;
-    font-size: 8pt;
+    font-size: 0.9em;
     background-color: rgba(0, 0, 0, 0.5);
     position:relative; 
-    top: -1.25pt;
+    top: -0.05em;
     margin-right: 2pt;
     user-select: none;
     /* exclude from text selection */
@@ -917,8 +926,8 @@ class Truncated extends React.Component {
 
   componentDidMount() {
     this.stepper = setInterval(() => {
-      this.setState(s => Object.assign(s, { typingOffset: s.typingOffset + 4 }))
-    }, 10)
+      this.setState(s => Object.assign(s, { typingOffset: s.typingOffset + 16 }))
+    }, 5)
   }
 
   componentWillUnmount() {
@@ -1733,6 +1742,7 @@ function SidePanel(props) {
 
   const [trackMostLikly, setTrackMostLiklyInternal] = useState(window.localStorage.getItem("trackMostLikely") === "true");
   trackingState.setTrackMostLikely = setTrackMostLiklyInternal
+  trackingState.getTrackMostLikely = () => trackMostLikly
   const setTrackMostLikly = (value) => {
     setTrackMostLiklyInternal(value)
     window.localStorage.setItem("trackMostLikely", value)
@@ -1848,6 +1858,24 @@ const Toolbar = styled.div`
   .bottom {
     margin-top: 40pt;
   }
+
+  >a {
+    display: block;
+    margin: 4pt;
+    margin-left: 8pt;
+    font-size: 10pt;
+    color: black;
+
+    text-decoration: none;
+
+    @media screen and (max-width: 40em) {
+      display: none;
+    }
+
+    :hover {
+      text-decoration: underline;
+    }
+  }
 `
 
 const ButtonGroup = styled.div`
@@ -1903,7 +1931,7 @@ const FancyButton = styled.button`
     }
   }
 
-  @media (max-width: 40em) {
+  @media (max-width: 50em) {
     &.in-toolbar {
       position: absolute;
       bottom: 20pt;
@@ -2000,7 +2028,8 @@ function DecoderPanel(props) {
     return {
       "_finfalse": data.user_data && data.user_data.head && data.user_data.head.valid == "False",
       "_isRoot": data.root,
-      "_isDone": data.user_data && data.user_data.head && data.user_data.head.variable == "__done__"
+      "_isDone": data.user_data && data.user_data.head && data.user_data.head.variable == "__done__",
+      "_noUserData": !data.user_data || (data.user_data && data.user_data == "None")
     }
   }
 
@@ -2338,6 +2367,9 @@ class App extends React.Component {
       topMenuOpen: false,
       // true by default, false if local storage is set
       simpleMode: window.localStorage.getItem("simple-mode") == "true",
+
+      /* show code screenshot overlay */
+      codeScreenshot: false
     }
   }
 
@@ -2426,11 +2458,28 @@ class App extends React.Component {
 
     // when document browser scale is changed
     window.addEventListener("resize", ResizeObservers.notify)
+    
+    this.onKeyDown = this.onKeyDown.bind(this)
+    window.addEventListener("keydown", this.onKeyDown)
+  }
+
+  onKeyDown(event) {
+    /* On R*/
+    if (event.keyCode == 82 && event.ctrlKey) {
+      this.onRun();
+    }
+    /* On Escape */
+    if (event.keyCode == 27) {
+      LMQLProcess.kill();
+    }
   }
 
   componentWillUnmount() {
     LMQLProcess.remove("render", this)
     LMQLProcess.remove("status", this.onStatus.bind(this))
+    
+    window.removeEventListener("resize", ResizeObservers.notify)
+    window.removeEventListener("keydown", this.onKeyDown)
   }
 
   onExportState() {
@@ -2481,13 +2530,34 @@ class App extends React.Component {
     this.setMostLikelyNode(node)
   };
 
+  getCodeScreenshotInput() {
+    let modelResult = null;
+    if (trackingState.getTrackMostLikely()) {
+      modelResult = reconstructTaggedModelResult([this.state.mostLikelyNode]);
+    } else {
+      modelResult = reconstructTaggedModelResult(this.state.selectedNodes);
+    }
+    return {
+      "code": persistedState.getItem("lmql-editor-contents"),
+      "model_result": modelResult,
+    }
+  }
+
+  onCodeScreenshot() {
+    this.setState({ codeScreenshot: true })
+  }
+
+  hideCodeScreenshot() {
+    this.setState({ codeScreenshot: false })
+  }
+
   render() {
     trackingState.setSelectedNode = n => {
       this.onSelectNode(n)
       this.state.selectedNodeTrigger.trigger(n)
     };
 
-    const simpleModeClassName = this.state.simpleMode ? "" : "simple-mode";
+    const simpleModeClassName = (this.state.simpleMode && displayState.mode != "embed") ? "" : "simple-mode";
 
     return (
       <ContentContainer className={this.state.graphLayout ? 'graph-layout' : ''}>
@@ -2497,6 +2567,8 @@ class App extends React.Component {
             LMQL Playground
           </Title>
           {configuration.DEMO_MODE && <FancyButton className="in-toolbar" onClick={() => ExploreState.setVisibility(true)}><ExploreIc/> Explore LMQL</FancyButton>}
+          {window.location.hostname.includes("lmql.ai") && <a href={"https://docs.lmql.ai/en/latest/quickstart.html"} target="_blank" rel="noreferrer" className="hidden-on-small">
+          Install LMQL Locally </a>}
           <Spacer />
           {/* show tooltip with build time */}
           {/* trigger button */}
@@ -2517,8 +2589,8 @@ class App extends React.Component {
               {configuration.BROWSER_MODE && <li onClick={() => OpenAICredentialsState.setOpen(true)}>
               <BsKeyFill/> OpenAI Credentials
               </li>}
-              {configuration.DEV_MODE && <li onClick={() => this.onExportState()}><BsFileArrowDownFill/> Export State
-              </li>}
+              {configuration.DEV_MODE && <li onClick={() => this.onExportState()}><BsFileArrowDownFill/> Export State</li>}
+              {configuration.DEV_MODE && <li onClick={() => this.onCodeScreenshot()}><BsFillCameraFill/> Code Screenshot</li>}
               <li>
                 <a href="https://github.com/eth-sri/lmql" disabled target="_blank" rel="noreferrer"><BsGithub/>LMQL on Github</a>
               </li>
@@ -2544,6 +2616,7 @@ class App extends React.Component {
         </Row>
         <OpenAICredentials />
         {configuration.DEMO_MODE && displayState.mode != "embed" && <Explore />}
+        {this.state.codeScreenshot && <CodeScreenshot hide={this.hideCodeScreenshot.bind(this)} {...this.getCodeScreenshotInput()}/>}
       </ContentContainer>
     );
   }
