@@ -173,6 +173,7 @@ function layoutDecoderGraph(cy) {
         roots: rootNodes,
         visit: (node, edge, previous, i, depth) => {
             if (node.hasClass("compound")) return;
+            if (node.data("_noUserData")) return;
 
             if (depth >= mostLikelyDepth) {
                 if (depth > mostLikelyDepth) {
@@ -279,6 +280,18 @@ function initDecoderGraphCy(element) {
                     'border-width': '0px',
                 }
             },
+            {
+                selector: 'node[_noUserData]',
+                style: {
+                    // light blue
+                    'background-color': 'rgb(249, 209, 248)',
+                    'background-opacity': 0.7,
+                    'color': 'white',
+                    'border-color': 'red',
+                    'border-width': '0px',
+                    'opacity': 0.4
+                }
+            },
             {
                 selector: 'node.inactive',
                 style: {
@@ -567,7 +580,11 @@ export function DecoderGraph(props) {
                             target: to
                         }
                     };
-                    cy.add({group: "edges", ...e})
+                    try {
+                        cy.add({group: "edges", ...e})
+                    } catch (e) {
+                        console.error("error adding edge", e)
+                    }
                 }))
 
                 let mostLikely = layoutDecoderGraph(cy)
@@ -588,11 +605,23 @@ export function DecoderGraph(props) {
         const renderer = {
             add_result: (output) => {
                 if (output.type == "decoder-graph-state") {
-                    // persist decoder graph in local state
-                    const raw = JSON.stringify(output.data)
-                    setRawGraphData(raw)
-                    persistedState.setItem("decoder-graph", raw, onPersistedGraphChange)
-                    setCyData(output.data)
+                    setCyData(cyData => {
+                        // persist decoder graph in local state
+                        let updated = Object.assign({nodes: [], edges: []}, cyData || {})
+                        let nodes = {}
+                        updated.nodes.forEach(n => nodes[n.id] = n)
+                        output.data.nodes.forEach(n => nodes[n.id] = n)
+                        updated.nodes = Array.from(Object.values(nodes))
+                        
+                        updated.edges = Array.from([...updated.edges, ...output.data.edges])
+                        
+                        // console.log("updating", output.data.nodes.length)
+
+                        const raw = JSON.stringify(updated)
+                        setRawGraphData(raw)
+                        persistedState.queueSetItem("decoder-graph", raw, onPersistedGraphChange)
+                        return updated
+                    })
                 } else {
                     // nop in this component
                 }
