@@ -1,3 +1,5 @@
+import {queries} from "./queries";
+
 function getSnippet(allow_snippet=true) {
   const url = window.location.href
   if (url.indexOf('?') === -1) {
@@ -16,6 +18,7 @@ function getSnippet(allow_snippet=true) {
 export const displayState = {
   // if ?embed, mode = embed, else playground
   mode: getSnippet(false) ? 'embed' : 'playground',
+  preloaded: getSnippet(true) ? true : false,
   embedFile: null
 }
 
@@ -24,6 +27,8 @@ class PersistedState {
       this.items = {}
       this.listeners = {}
       this.restore()
+
+      this.saveQueue = {}
     }
   
     persist(k) {
@@ -42,12 +47,42 @@ class PersistedState {
       // check for snippet or embed
       let snippetFile = getSnippet()
       if (snippetFile) {
+        let load_as_json = true;
+
         console.log("loading snippet from", snippetFile)
+        let matches = queries.filter(c => c.queries.find(q => q.state == "precomputed/" + snippetFile + ".json"))
+        if (matches.length === 1) {
+          let q = matches[0].queries.find(q => q.state == "precomputed/" + snippetFile + ".json")
+          snippetFile = q.state
+        }
+
+        // check for github gist
+        if (snippetFile.startsWith("gist:")) {
+          // will be of this form gist:lbeurerkellner/9e203be6f9120dae2515c43517bc1aee/raw/test.lmql
+          try {
+            let parts = snippetFile.split(":")
+            let user = parts[1].split("/")[0]
+            let gist = parts[1].split("/")[1]
+            let file = parts[1].split("/")[3]
+            snippetFile = `https://gist.githubusercontent.com/${user}/${gist}/raw/${file}`
+
+            load_as_json = file.endsWith(".json")
+          } catch (e) {
+            console.error("error parsing github gist URL", e)
+            return;
+          }
+        }
+
         fetch(snippetFile).then(r => r.text()).then(data => {
-          console.log(data)
           // remove ? from url
-          window.history.replaceState({}, document.title, window.location.pathname);
+          if (!snippetFile.includes("gist.github")) {
+            window.history.replaceState({}, document.title, window.location.pathname);
+          }
+          // actually load data
           if (data) {
+            if (!load_as_json) {
+              data = JSON.stringify({ "lmql-editor-contents": data , "decoder-graph":"{\"nodes\":[],\"edges\":[]}"})
+            }
             displayState.embedFile = snippetFile
             this.load(data)
           }
@@ -108,12 +143,22 @@ class PersistedState {
         })
       }
     }
+
+    queueSetItem(key, value, exclude_listener=null) {
+      if (this.saveQueue[key]) {
+        clearTimeout(this.saveQueue[key]);
+      }
+      this.saveQueue[key] = setTimeout(() => {
+        this.setItem(key, value, exclude_listener);
+      })
+    }
   }
   
 export const persistedState = new PersistedState();
 
 export const trackingState = {
   setTrackMostLikely: () => { },
+  getTrackMostLikely: () => false,
   setSelectedNode: () => { },
 }
 
